"""
This file is part of the openPMD-viewer.

It defines a number of helper functions that are used in main.py

Copyright 2015-2017, openPMD-viewer contributors
Authors: Remi Lehe, Richard Pausch
License: 3-Clause-BSD-LBNL
"""

import copy
import math
import numpy as np
from .numba_wrapper import jit
from .data_order import RZorder, order_error_msg

def sanitize_slicing(slice_across, slice_relative_position):
    """
    Return standardized format for `slice_across` and `slice_relative_position`:
    - either `slice_across` and `slice_relative_position` are both `None` (no slicing)
    - or `slice_across` and `slice_relative_position` are both lists,
    with the same number of elements

    Parameters
    ----------
    slice_relative_position : float, or list of float, or None

    slice_across : str, or list of str, or None
       Direction(s) across which the data should be sliced
    """
    # Skip None and empty lists
    if slice_across is None or slice_across == []:
        return None, None

    # Convert to lists
    if not isinstance(slice_across, list):
        slice_across = [slice_across]
    if slice_relative_position is None:
        slice_relative_position = [0]*len(slice_across)
    if not isinstance(slice_relative_position, list):
        slice_relative_position = [slice_relative_position]
    # Check that the length are matching
    if len(slice_across) != len(slice_relative_position):
        raise ValueError(
            'The argument `slice_relative_position` is erroneous: \nIt should have'
            'the same number of elements as `slice_across`.')

    # Return a copy. This is because the rest of the `openPMD-viewer` code
    # sometimes modifies the objects returned by `sanitize_slicing`.
    # Using a copy avoids directly modifying objects that the user may pass
    # to this function (and live outside of openPMD-viewer, e.g. directly in
    # a user's notebook)
    return copy.copy(slice_across), copy.copy(slice_relative_position)

def apply_selection(iteration, data_reader, data_list,
                    select, species, extensions):
    """
    Select the elements of each particle quantities in data_list,
    based on the selection rules in `select`

    Parameters
    ----------
    iteration: int
        The iteration at which to apply the selection

    data_reader: a DataReader object
        Contains the method that read particle data

    data_list: list of 1darrays
        A list of arrays with one element per macroparticle, that represent
        different particle quantities

    select: dict
        A dictionary of rules to select the particles
        'x' : [-4., 10.]   (Particles having x between -4 and 10)
        'ux' : [-0.1, 0.1] (Particles having ux between -0.1 and 0.1 mc)
        'uz' : [5., None]  (Particles with uz above 5 mc)

    species: string
       Name of the species being requested

    extensions: list of strings
        The extensions that the current OpenPMDTimeSeries complies with

    Returns
    -------
    A list of 1darrays that correspond to data_list, but were only the
    macroparticles that meet the selection rules are kept
    """
    # Create the array that determines whether the particle
    # should be selected or not.
    Ntot = len(data_list[0])
    select_array = np.ones(Ntot, dtype='bool')

    # Loop through the selection rules, and aggregate results in select_array
    for quantity in select.keys():
        q = data_reader.read_species_data(
            iteration, species, quantity, extensions)
        # Check lower bound
        if select[quantity][0] is not None:
            select_array = np.logical_and(
                select_array,
                q > select[quantity][0])
        # Check upper bound
        if select[quantity][1] is not None:
            select_array = np.logical_and(
                select_array,
                q < select[quantity][1])

    # Use select_array to reduce each quantity
    for i in range(len(data_list)):
        if len(data_list[i]) > 1:  # Do not apply selection on scalar records
            data_list[i] = data_list[i][select_array]

    return(data_list)


def try_array( L ):
    """
    Attempt to convert L to a single array.
    """
    try:
        # Stack the arrays
        return np.stack( L, axis=0 )
    except ValueError:
        # Do not stack
        return L


def fit_bins_to_grid( hist_size, grid_size, grid_range ):
    """
    Given a tentative number of bins `hist_size` for a histogram over
    the range `grid_range`, return a modified number of bins `hist_size`
    and a modified range `hist_range` so that the spacing of the histogram
    bins is an integer multiple (or integer divisor) of the grid spacing.

    Parameters:
    ----------
    hist_size: integer
        The number of bins in the histogram along the considered direction

    grid_size: integer
        The number of cells in the grid

    grid_range: list of floats (in)
        The extent of the grid

    Returns:
    --------
    hist_size: integer
        The new number of bins

    hist_range: list of floats
        The new range of the histogram
    """
    # The new histogram range is the same as the grid range
    hist_range = grid_range

    # Calculate histogram tentative spacing, and grid spacing
    hist_spacing = ( hist_range[1] - hist_range[0] ) * 1. / hist_size
    grid_spacing = ( grid_range[1] - grid_range[0] ) * 1. / grid_size

    # Modify the histogram spacing, so that either:
    if hist_spacing >= grid_spacing:
        # - The histogram spacing is an integer multiple of the grid spacing
        hist_spacing = int( hist_spacing / grid_spacing ) * grid_spacing
    else:
        # - The histogram spacing is an integer divisor of the grid spacing
        hist_spacing = grid_spacing / int( grid_spacing / hist_spacing )

    # Get the corresponding new number of bins, and the new range
    hist_size = int( ( hist_range[1] - hist_range[0] ) / hist_spacing )
    hist_range[1] = hist_range[0] + hist_size * hist_spacing

    return( hist_size, hist_range )


def combine_cylindrical_components( Fr, Ft, theta, coord, info ):
    """
    Calculate the catesian field Fx or Fy,
    from the cylindrical components Fr and Ft.

    Parameters:
    -----------
    Fr, Ft: 3darrays or 2Darrays (depending on whether `theta` is None)
        Contains the value of the fields
    theta: float or None
        Indicates the angle of the plane in which Fr and Ft where taken
    coord: string
        Either 'x' or 'y' ; indicates which component to calculate
    info: FieldMetaInformation object
        Contains info on the coordinate system
    """
    if theta is not None:
        if coord == 'x':
            F = np.cos(theta) * Fr - np.sin(theta) * Ft
        elif coord == 'y':
            F = np.sin(theta) * Fr + np.cos(theta) * Ft
        # Revert the sign below the axis
        if info.axes[0] == 'r':
            F[ : int(F.shape[0]/2) ] *= -1
        elif (F.ndim == 2) and (info.axes[1] == 'r'):
            F[ : , : int(F.shape[1]/2) ] *= -1
    else:
        # Fr, Ft are 3Darrays, info corresponds to Cartesian data
        assert (Fr.ndim == 3) and (Ft.ndim == 3)

        # Calculate cos(theta) and sin(theta) in the transverse Cartesian plane
        # while avoiding divisions by 0
        r = np.sqrt( info.x[:,np.newaxis]**2 + info.y[np.newaxis,:]**2 )
        inv_r = 1./np.where( r!=0, r, 1. )
        # The value `1.`` is a placeholder in the above (to avoid division by 0)
        # The lines below replace this placeholder value.
        cos = np.where( r!=0, info.x[:,np.newaxis]*inv_r, 1. )
        sin = np.where( r!=0, info.y[np.newaxis,:]*inv_r, 0. )
        if coord == 'x':
            F = cos[:,:,np.newaxis] * Fr - sin[:,:,np.newaxis] * Ft
        elif coord == 'y':
            F = sin[:,:,np.newaxis] * Fr + cos[:,:,np.newaxis] * Ft

    return F

@jit
def histogram_cic_1d( q1, w, nbins, bins_start, bins_end ):
    """
    Return an 1D histogram of the values in `q1` weighted by `w`,
    consisting of `nbins` evenly-spaced bins between `bins_start`
    and `bins_end`. Contribution to each bins is determined by the
    CIC weighting scheme (i.e. linear weights).
    """
    # Define various scalars
    bin_spacing = (bins_end-bins_start)/nbins
    inv_spacing = 1./bin_spacing
    n_ptcl = len(w)

    # Allocate array for histogrammed data
    hist_data = np.zeros( nbins, dtype=np.float64 )

    # Go through particle array and bin the data
    for i in range(n_ptcl):
        # Calculate the index of lower bin to which this particle contributes
        q1_cell = (q1[i] - bins_start) * inv_spacing
        i_low_bin = int( math.floor( q1_cell ) )
        # Calculate corresponding CIC shape and deposit the weight
        S_low = 1. - (q1_cell - i_low_bin)
        if (i_low_bin >= 0) and (i_low_bin < nbins):
            hist_data[ i_low_bin ] += w[i] * S_low
        if (i_low_bin + 1 >= 0) and (i_low_bin + 1 < nbins):
            hist_data[ i_low_bin + 1 ] += w[i] * (1. - S_low)

    return( hist_data )


@jit
def histogram_cic_2d( q1, q2, w,
    nbins_1, bins_start_1, bins_end_1,
    nbins_2, bins_start_2, bins_end_2 ):
    """
    Return an 2D histogram of the values in `q1` and `q2` weighted by `w`,
    consisting of `nbins_1` bins in the first dimension and `nbins_2` bins
    in the second dimension.
    Contribution to each bins is determined by the
    CIC weighting scheme (i.e. linear weights).
    """
    # Define various scalars
    bin_spacing_1 = (bins_end_1-bins_start_1)/nbins_1
    inv_spacing_1 = 1./bin_spacing_1
    bin_spacing_2 = (bins_end_2-bins_start_2)/nbins_2
    inv_spacing_2 = 1./bin_spacing_2
    n_ptcl = len(w)

    # Allocate array for histogrammed data
    hist_data = np.zeros( (nbins_1, nbins_2), dtype=np.float64 )

    # Go through particle array and bin the data
    for i in range(n_ptcl):

        # Calculate the index of lower bin to which this particle contributes
        q1_cell = (q1[i] - bins_start_1) * inv_spacing_1
        q2_cell = (q2[i] - bins_start_2) * inv_spacing_2
        i1_low_bin = int( math.floor( q1_cell ) )
        i2_low_bin = int( math.floor( q2_cell ) )

        # Calculate corresponding CIC shape and deposit the weight
        S1_low = 1. - (q1_cell - i1_low_bin)
        S2_low = 1. - (q2_cell - i2_low_bin)
        if (i1_low_bin >= 0) and (i1_low_bin < nbins_1):
            if (i2_low_bin >= 0) and (i2_low_bin < nbins_2):
                hist_data[ i1_low_bin, i2_low_bin ] += w[i]*S1_low*S2_low
            if (i2_low_bin+1 >= 0) and (i2_low_bin+1 < nbins_2):
                hist_data[ i1_low_bin, i2_low_bin+1 ] += w[i]*S1_low*(1.-S2_low)
        if (i1_low_bin+1 >= 0) and (i1_low_bin+1 < nbins_1):
            if (i2_low_bin >= 0) and (i2_low_bin < nbins_2):
                hist_data[ i1_low_bin+1, i2_low_bin ] += w[i]*(1.-S1_low)*S2_low
            if (i2_low_bin+1 >= 0) and (i2_low_bin+1 < nbins_2):
                hist_data[ i1_low_bin+1, i2_low_bin+1 ] += w[i]*(1.-S1_low)*(1.-S2_low)

    return( hist_data )


@jit
def construct_3d_from_circ( F3d, Fcirc, x_array, y_array, modes,
    nx, ny, nz, nr, nmodes, inv_dr, rmax, coord_order): 
    """
    Reconstruct the field from a quasi-cylindrical simulation (`Fcirc`), as
    a 3D cartesian array (`F3d`).
    """
    for ix in range(nx):
        x = x_array[ix]
        for iy in range(ny):
            y = y_array[iy]
            r = np.sqrt( x**2 + y**2 )
            ir = nr - 1 - int( (rmax - r) * inv_dr + 0.5 )

            # Handle out-of-bounds
            if ir < 0:
                ir = 0
            if ir >= nr:
                ir = nr-1

            # Calculate linear projection from ir and ir-1
            if ir>0:
                s0 = ir + 0.5 - r* inv_dr
                s1 = 1. - s0
                if coord_order is RZorder.mrz:
                    Fcirc_proj = s1*Fcirc[:, ir, :] + s0*Fcirc[:, ir-1, :]
                elif coord_order is RZorder.mzr:
                    Fcirc_proj = s1*Fcirc[:, :, ir] + s0*Fcirc[:, :, ir-1]
                else:
                    raise Exception(order_error_msg)
            else:
                if coord_order is RZorder.mrz:
                    Fcirc_proj = Fcirc[:, ir, :]
                elif coord_order is RZorder.mzr:
                    Fcirc_proj = Fcirc[:, :, ir]
                else:
                    raise Exception(order_error_msg)

            # Loop over all modes and recontruct data
            if r == 0:
                expItheta = 1. + 0.j
            else:
                expItheta = (x+1.j*y)/r

            for im in range(nmodes):
                mode = modes[im]
                if mode==0:
                    F3d[ix, iy, :] += Fcirc_proj[0, :]
                else:
                    cos = (expItheta**mode).real
                    sin = (expItheta**mode).imag
                    F3d[ix, iy, :] += Fcirc_proj[2*mode-1,:]*cos + \
                        Fcirc_proj[2*mode,:]*sin



"""
This file is part of the openPMD-viewer.

It defines a function that can read standard parameters from an openPMD file.

Copyright 2020, openPMD-viewer contributors
Authors: Axel Huebl
License: 3-Clause-BSD-LBNL
"""

import numpy as np
from .utilities import join_infile_path


def read_openPMD_params(series, iteration, extract_parameters=True):
    """
    Extract the time and some openPMD parameters from a file

    Parameter
    ---------
    series: openpmd_api.Series
        An open, readable openPMD-api series object

    iteration: integer
        Iteration from which parameters should be extracted

    extract_parameters: bool, optional
        Whether to extract all parameters or only the time
        (Function execution is faster when extract_parameters is False)

    Returns
    -------
    A tuple with:
    - A float corresponding to the time of this iteration in SI units
    - A dictionary containing several parameters, such as the geometry, etc.
      When extract_parameters is False, the second argument returned is None.
    """
    it = series.iterations[iteration]

    # extract the time
    if callable(it.time):  # prior to openPMD-api 0.13.0
        t = it.time() * it.time_unit_SI()
    else:
        t = it.time * it.time_unit_SI

    # If the user did not request more parameters, close file and exit
    if not extract_parameters:
        return t, None

    # Otherwise, extract the rest of the parameters
    params = {}

    # Find out supported openPMD extensions claimed by this file
    # note: a file might implement multiple extensions
    known_extensions = {'ED-PIC': np.uint32(1)}
    bitmask_all_extensions = series.openPMD_extension
    params['extensions'] = []
    for extension, bitmask in known_extensions.items():
        # This uses a bitmask to identify activated extensions
        # efficiently in static programming languages via
        # a single attribute and a binary AND (&) operation.
        # Standard: https://git.io/vwnMw
        # Bitmasks: https://en.wikipedia.org/wiki/Mask_%28computing%29
        if bitmask_all_extensions & bitmask == bitmask:
            params['extensions'].append(extension)

    # Find out whether fields are present and extract their metadata
    fields_available = len(it.meshes) > 0
    if fields_available:
        params['avail_fields'] = []
        params['fields_metadata'] = {}

        # Loop through the available fields
        for field_name, field in it.meshes.items():
            metadata = {}
            metadata['geometry'] = field.get_attribute('geometry')
            metadata['axis_labels'] = field.axis_labels

            # Swap the order of the labels if the code that wrote the HDF5 file
            # was Fortran order (i.e. reverse order with respect to Python)
            if field.get_attribute('dataOrder') == 'F':
                metadata['axis_labels'] = metadata['axis_labels'][::-1]
            # Check whether the field is a vector or a scalar
            if field.scalar:
                metadata['type'] = 'scalar'
                components = []
            else:
                metadata['type'] = 'vector'
                components = [comp for comp, _ in field.items()]
            # Register available components
            metadata['avail_components'] = components
            # Check the number of modes
            if metadata['geometry'] == "thetaMode":
                # simply check first record component
                field_component = next(field.items())[1]
                Nm = field_component.shape[0]
                metadata['avail_circ_modes'] = ['all'] + \
                    [str(m) for m in range(int(Nm / 2) + 1)]
            # Check if this a 1d, 2d or 3d Cartesian
            elif metadata['geometry'] == "cartesian":
                dim = len(metadata['axis_labels'])
                if dim == 1:
                    metadata['geometry'] = "1dcartesian"
                elif dim == 2:
                    metadata['geometry'] = "2dcartesian"
                elif dim == 3:
                    metadata['geometry'] = "3dcartesian"
                metadata['avail_circ_modes'] = []

            params['avail_fields'].append( field_name )
            params['fields_metadata'][field_name] = metadata

    else:
        params['avail_fields'] = None

    # Find out whether particles are present, and if yes of which species
    particles_available = len(it.particles) > 0
    if particles_available:
        # Particles are present ; extract the species
        params['avail_species'] = list(it.particles)
        # dictionary with list of record components for each species
        record_components = {}
        # Go through all species
        # TODO: I did this more elegant in ParaView... for later
        for species_name, species in it.particles.items():
            record_components[species_name] = []

            # Go through all the particle records of this species
            for record_name, record in species.items():
                # Skip the particlePatches, which are not used here.
                # API should filter this...
                # if record_name == 'particlePatches':
                #     continue
                if record.scalar:
                    # Add the name of the scalar record
                    record_components[species_name]. \
                        append(record_name)
                else:
                    # Add each component of the vector record
                    for compo_name in list(record):
                        record_components[species_name]. \
                            append(join_infile_path(record_name, compo_name))
            # Simplify the name of some standard openPMD records
            record_components[species_name] = \
                simplify_record(record_components[species_name])
        params['avail_record_components'] = record_components
        # deprecated
        first_species_name = next(iter(params['avail_species']))
        params['avail_ptcl_quantities'] = \
            record_components[first_species_name]
    else:
        # Particles are absent
        params['avail_species'] = None
        params['avail_record_components'] = None
        # deprecated
        params['avail_ptcl_quantities'] = None

    return t, params


def simplify_record(record_comps):
    """
    Replace the names of some standard record by shorter names

    Parameter
    ---------
    record_comps: a list of strings
        A list of available particle record components

    Returns
    -------
    A list with shorter names, where applicable
    """
    # Replace the names of the positions
    if ('position/x' in record_comps) and ('positionOffset/x' in record_comps):
        record_comps.remove('position/x')
        record_comps.remove('positionOffset/x')
        record_comps.append('x')
    if ('position/y' in record_comps) and ('positionOffset/y' in record_comps):
        record_comps.remove('position/y')
        record_comps.remove('positionOffset/y')
        record_comps.append('y')
    if ('position/z' in record_comps) and ('positionOffset/z' in record_comps):
        record_comps.remove('position/z')
        record_comps.remove('positionOffset/z')
        record_comps.append('z')

    # Replace the names of the momenta
    if 'momentum/x' in record_comps:
        record_comps.remove('momentum/x')
        record_comps.append('ux')
    if 'momentum/y' in record_comps:
        record_comps.remove('momentum/y')
        record_comps.append('uy')
    if 'momentum/z' in record_comps:
        record_comps.remove('momentum/z')
        record_comps.append('uz')

    # Replace the name for 'weights'
    if 'weighting' in record_comps:
        record_comps.remove('weighting')
        record_comps.append('w')

    return record_comps



"""
This file is part of the openPMD-viewer.

It defines a set of helper data and functions which
are used by the other files.

Copyright 2015-2016, openPMD-viewer contributors
Authors: Remi Lehe, Axel Huebl
License: 3-Clause-BSD-LBNL
"""
import os
import h5py
import numpy as np


def list_files(path_to_dir):
    """
    Return a list of the hdf5 files in this directory,
    and a list of the corresponding iterations

    Parameter
    ---------
    path_to_dir : string
        The path to the directory where the hdf5 files are.

    Returns
    -------
    A tuple with:
    - an array of integers which correspond to the iteration of each file
    - a dictionary that matches iterations to the corresponding filename
    """
    # group based encoding?
    is_single_file = os.path.isfile(path_to_dir)

    if is_single_file:
        all_files = [path_to_dir]
    else:
        # Find all the files in the provided directory
        all_files = os.listdir(path_to_dir)

    # Select the hdf5 files, and fill dictionary of correspondence
    # between iterations and files
    iteration_to_file = {}
    for filename in all_files:
        # Use only the name that end with .h5 or .hdf5
        if filename.endswith('.h5') or filename.endswith('.hdf5'):
            if is_single_file:
                full_name = filename
            else:
                full_name = os.path.join(
                    os.path.abspath(path_to_dir), filename)
            # extract all iterations from hdf5 file
            f = h5py.File(full_name, 'r')
            iterations = list(f['/data'].keys())
            f.close()
            # Add iterations to dictionary
            for key_iteration in iterations:
                iteration_to_file[ int(key_iteration) ] = full_name

    # Extract iterations and sort them
    iterations = np.array( sorted( list( iteration_to_file.keys() ) ) )

    return iterations, iteration_to_file


def is_scalar_record(record):
    """
    Determine whether a record is a scalar record or a vector record

    Parameter
    ---------
    record: an h5py Dataset or an h5py Group

    Return
    ------
    A boolean indicating whether the record is scalar
    """
    scalar = False
    if 'value' in record.attrs:
        scalar = True
    elif isinstance(record, h5py.Dataset):
        scalar = True

    return(scalar)


def get_data(dset, i_slice=None, pos_slice=None, output_type=None):
    """
    Extract the data from a (possibly constant) dataset
    Slice the data according to the parameters i_slice and pos_slice

    Parameters:
    -----------
    dset: an h5py.Dataset or h5py.Group (when constant)
        The object from which the data is extracted

    pos_slice: int or list of int, optional
        Slice direction(s).
        When None, no slicing is performed

    i_slice: int or list of int, optional
       Indices of slices to be taken.

    output_type: a numpy type
       The type to which the returned array should be converted

    Returns:
    --------
    An np.ndarray (non-constant dataset) or a single double (constant dataset)
    """
    # For back-compatibility: Convert pos_slice and i_slice to
    # single-element lists if they are not lists (e.g. float
    # and int respectively).
    if pos_slice is not None and not isinstance(pos_slice, list):
        pos_slice = [pos_slice]
    if i_slice is not None and not isinstance(i_slice, list):
        i_slice = [i_slice]
    # Case of a constant dataset
    if isinstance(dset, h5py.Group):
        shape = dset.attrs['shape']
        # Restrict the shape if slicing is enabled
        if pos_slice is not None:
            shape = [ x for index, x in enumerate(shape) if
                      index not in pos_slice ]
        # Create the corresponding dataset
        data = dset.attrs['value'] * np.ones(shape)

    # Case of a non-constant dataset
    elif isinstance(dset, h5py.Dataset):
        if pos_slice is None:
            data = dset[...]
        else:
            # Get largest element of pos_slice
            max_pos = max(pos_slice)
            # Create list of indices list_index of type
            # [:, :, :, ...] where Ellipsis starts at max_pos + 1
            list_index = [np.s_[:]] * (max_pos + 2)
            list_index[max_pos + 1] = np.s_[...]
            # Fill list_index with elements of i_slice
            for count, dir_index in enumerate(pos_slice):
                list_index[dir_index] = i_slice[count]
            # Convert list_index into a tuple
            tuple_index = tuple(list_index)
            # Slice dset according to tuple_index
            data = dset[tuple_index]

    # Convert to the right type
    if (output_type is not None) and (data.dtype != output_type):
        data = data.astype( output_type )
    # Scale by the conversion factor
    if np.issubdtype(data.dtype, np.floating) or \
        np.issubdtype(data.dtype, np.complexfloating):
        if dset.attrs['unitSI'] != 1.0:
            data *= dset.attrs['unitSI']

    return(data)


def get_shape(dset):
    """
    Extract the shape of a (possibly constant) dataset

    Parameters:
    -----------
    dset: an h5py.Dataset or h5py.Group (when constant)
        The object whose shape is extracted

    Returns:
    --------
    A tuple corresponding to the shape
    """
    # Case of a constant dataset
    if isinstance(dset, h5py.Group):
        shape = dset.attrs['shape']
    # Case of a non-constant dataset
    elif isinstance(dset, h5py.Dataset):
        shape = dset.shape

    return(shape)


def join_infile_path(*paths):
    """
    Join path components using '/' as separator.
    This method is defined as an alternative to os.path.join, which uses '\\'
    as separator in Windows environments and is therefore not valid to navigate
    within data files.

    Parameters:
    -----------
    *paths: all strings with path components to join

    Returns:
    --------
    A string with the complete path using '/' as separator.
    """
    # Join path components
    path = '/'.join(paths)
    # Correct double slashes, if any is present
    path = path.replace('//', '/')

    return path



"""
This file is part of the openPMD-viewer.

It defines a wrapper around numba.

Copyright 2019, openPMD-viewer contributors
Author: Remi Lehe
License: 3-Clause-BSD-LBNL
"""
import warnings

try:
    # Import jit decorator from numba
    import numba
    numba_installed = True
    jit = numba.njit(cache=True)

except ImportError:
    numba_installed = False
    # Create dummy decorator: warns about installing numba when calling
    # the decorated function.
    def jit(f):
        def decorated_f(*args, **kwargs):
            warnings.warn(
                '\nOne of the functions called by openPMD-viewer ' +\
                '(%s)\n' %f.__name__ +\
                'could have been faster if `numba` had been installed.\n' +\
                'Please consider installing `numba` (e.g. `pip install numba`)')
            return f(*args, **kwargs)
        return decorated_f



"""
This file is part of the openPMD-viewer.

It defines functions that can read the fields from an HDF5 file.

Copyright 2020, openPMD-viewer contributors
Author: Axel Huebl
License: 3-Clause-BSD-LBNL
"""

import numpy as np

from .utilities import get_data
from ...data_order import RZorder, order_error_msg
from openpmd_viewer.openpmd_timeseries.field_metainfo import FieldMetaInformation
from openpmd_viewer.openpmd_timeseries.utilities import construct_3d_from_circ


def read_field_cartesian( series, iteration, field_name, component_name,
                          axis_labels, slice_relative_position, slice_across ):
    """
    Extract a given field from a file in the openPMD format,
    when the geometry is cartesian (1d, 2d or 3d).

    Parameters
    ----------
    series: openpmd_api.Series
        An open, readable openPMD-api series object

    iteration: integer
        Iteration from which parameters should be extracted

    field_name : string, optional
       Which field to extract

    component_name : string, optional
       Which component of the field to extract

    axis_labels: list of strings
       The name of the dimensions of the array (e.g. ['x', 'y', 'z'])

    slice_across : list of str or None
       Direction(s) across which the data should be sliced
       Elements can be:
         - 1d: 'z'
         - 2d: 'x' and/or 'z'
         - 3d: 'x' and/or 'y' and/or 'z'
       Returned array is reduced by 1 dimension per slicing.

    slice_relative_position : list of float or None
       Number(s) between -1 and 1 that indicate where to slice the data,
       along the directions in `slice_across`
       -1 : lower edge of the simulation box
       0 : middle of the simulation box
       1 : upper edge of the simulation box

    Returns
    -------
    A tuple with
       F : a ndarray containing the required field
       info : a FieldMetaInformation object
       (contains information about the grid; see the corresponding docstring)
    """
    it = series.iterations[iteration]

    # Extract the dataset and corresponding group
    field = it.meshes[field_name]
    if field.scalar:
        component = next(field.items())[1]
    else:
        component = field[component_name]

    # Dimensions of the grid
    shape = component.shape
    # FIXME here and in h5py reader, we need to invert the order on 'F'
    grid_spacing = field.grid_spacing
    global_offset = field.grid_global_offset
    grid_unit_SI = field.grid_unit_SI
    grid_position = component.position
    time = (it.time + field.time_offset) * it.time_unit_SI

    # Slice selection
    #   TODO put in general utilities
    if slice_across is not None:
        # Get the integer that correspond to the slicing direction
        list_slicing_index = []
        list_i_cell = []
        for count, slice_across_item in enumerate(slice_across):
            slicing_index = axis_labels.index(slice_across_item)
            list_slicing_index.append(slicing_index)
            # Number of cells along the slicing direction
            n_cells = shape[ slicing_index ]
            # Index of the slice (prevent stepping out of the array)
            i_cell = int( 0.5 * (slice_relative_position[count] + 1.) * n_cells )
            i_cell = max( i_cell, 0 )
            i_cell = min( i_cell, n_cells - 1)
            list_i_cell.append(i_cell)

        # Remove metainformation relative to the slicing index
        # Successive pops starting from last coordinate to slice
        shape = [ x for index, x in enumerate(shape)
                  if index not in list_slicing_index ]
        grid_spacing = [ x for index, x in enumerate(grid_spacing)
                         if index not in list_slicing_index ]
        global_offset = [ x for index, x in enumerate(global_offset)
                          if index not in list_slicing_index ]
        axis_labels = [ x for index, x in enumerate(axis_labels)
                         if index not in list_slicing_index ]

        axes = { i: axis_labels[i] for i in range(len(axis_labels)) }
        # Extract data
        F = get_data( series, component, list_i_cell, list_slicing_index )
        info = FieldMetaInformation( axes, shape, grid_spacing, global_offset,
                grid_unit_SI, grid_position,
                time, iteration )
    else:
        F = get_data( series, component )
        axes = { i: axis_labels[i] for i in range(len(axis_labels)) }
        info = FieldMetaInformation( axes, F.shape,
            grid_spacing, global_offset,
            grid_unit_SI, grid_position,
            time, iteration )

    return F, info


def read_field_circ( series, iteration, field_name, component_name,
                     slice_relative_position, slice_across, m=0, theta=0.,
                     max_resolution_3d=None ):
    """
    Extract a given field from a file in the openPMD format,
    when the geometry is thetaMode

    Parameters
    ----------
    series: openpmd_api.Series
        An open, readable openPMD-api series object

    iteration: integer
        Iteration from which parameters should be extracted

    field_name : string, optional
       Which field to extract

    component_name : string, optional
       Which component of the field to extract

    m : int or string, optional
       The azimuthal mode to be extracted

    theta : float or None
       Angle of the plane of observation with respect to the x axis
       If `theta` is not None, then this function returns a 2D array
       corresponding to the plane of observation given by `theta` ;
       otherwise it returns a full 3D Cartesian array

    slice_across : list of str or None
       Direction(s) across which the data should be sliced
       Elements can be 'r' and/or 'z'
       Returned array is reduced by 1 dimension per slicing.

    slice_relative_position : list of float or None
       Number(s) between -1 and 1 that indicate where to slice the data,
       along the directions in `slice_across`
       -1 : lower edge of the simulation box
       0 : middle of the simulation box
       1 : upper edge of the simulation box

    max_resolution_3d : list of int or None
        Maximum resolution that the 3D reconstruction of the field (when
        `theta` is None) can have. The list should contain two values,
        e.g. `[200, 100]`, indicating the maximum longitudinal and transverse
        resolution, respectively. This is useful for performance reasons,
        particularly for 3D visualization.

    Returns
    -------
    A tuple with
       F : a 3darray or 2darray containing the required field,
           depending on whether `theta` is None or not
       info : a FieldMetaInformation object
       (contains information about the grid; see the corresponding docstring)
    """
    it = series.iterations[iteration]

    # Extract the dataset and corresponding group
    field = it.meshes[field_name]
    if field.scalar:
        component = next(field.items())[1]
    else:
        component = field[component_name]

    # Extract the metainformation
    #   FIXME here and in h5py reader, we need to invert the order on 'F' for
    #         grid spacing/offset/position

    coord_labels = {ii: coord for (ii, coord) in enumerate(field.axis_labels)}
    coord_label_str = ''.join(coord_labels.values())
    coord_label_str = 'm' + coord_label_str
    coord_order = RZorder[coord_label_str]
    if coord_order is RZorder.mrz:
        Nm, Nr, Nz = component.shape
        N_pair = (Nr, Nz)
    elif coord_order is RZorder.mzr:
        Nm, Nz, Nr = component.shape
        N_pair = (Nz, Nr)
    else:
        raise Exception(order_error_msg)
    time = (it.time + field.time_offset) * it.time_unit_SI

    # Nm, Nr, Nz = component.shape
    info = FieldMetaInformation( coord_labels, N_pair,
        field.grid_spacing, field.grid_global_offset,
        field.grid_unit_SI, component.position, time, iteration, thetaMode=True )

    # Convert to a 3D Cartesian array if theta is None
    if theta is None:

        # Get cylindrical info
        rmax = info.rmax
        inv_dr = 1./info.dr
        Fcirc = get_data( series, component )  # (Extracts all modes)
        if m == 'all':
            modes = [ mode for mode in range(0, int(Nm / 2) + 1) ]
        else:
            modes = [ m ]
        modes = np.array( modes, dtype='int' )
        nmodes = len(modes)

        # If necessary, reduce resolution of 3D reconstruction
        if max_resolution_3d is not None:
            max_res_lon, max_res_transv = max_resolution_3d
            if Nz > max_res_lon:
                # Calculate excess of elements along z
                excess_z = int(np.round(Nz/max_res_lon))
                # Preserve only one every excess_z elements
                if coord_order is RZorder.mrz:
                    Fcirc = Fcirc[:, :, ::excess_z]
                elif coord_order is RZorder.mzr:
                    Fcirc = Fcirc[:, ::excess_z, :]
                else:
                    raise Exception(order_error_msg)
                # Update info accordingly
                info.z = info.z[::excess_z]
                info.dz = info.z[1] - info.z[0]
            if Nr > max_res_transv/2:
                # Calculate excess of elements along r
                excess_r = int(np.round(Nr/(max_res_transv/2)))
                # Preserve only one every excess_r elements
                if coord_order is RZorder.mrz:
                    Fcirc = Fcirc[:, ::excess_r, :]
                elif coord_order is RZorder.mzr:
                    Fcirc = Fcirc[:, :, ::excess_r]
                else:
                    raise Exception(order_error_msg)
                # Update info and necessary parameters accordingly
                info.r = info.r[::excess_r]
                info.dr = info.r[1] - info.r[0]
                inv_dr = 1./info.dr
                # Update Nr after reducing radial resolution.
                if coord_order is RZorder.mrz:
                    Nr = Fcirc.shape[1]
                elif coord_order is RZorder.mzr:
                    Nr = Fcirc.shape[2]
                else:
                    raise Exception(order_error_msg)

        # Convert cylindrical data to Cartesian data
        info._convert_cylindrical_to_3Dcartesian()
        nx, ny, nz = len(info.x), len(info.y), len(info.z)
        F_total = np.zeros( (nx, ny, nz) )
        construct_3d_from_circ( F_total, Fcirc, info.x, info.y, modes,
            nx, ny, nz, Nr, nmodes, inv_dr, rmax, coord_order)

    else:

        # Extract the modes and recombine them properly
        if coord_order is RZorder.mrz:
            F_total = np.zeros( (2 * Nr, Nz ) )
        elif coord_order is RZorder.mzr:
            F_total = np.zeros( (Nz, 2 * Nr ) )
        else:
            raise Exception(order_error_msg)
        if m == 'all':
            # Sum of all the modes
            # - Prepare the multiplier arrays
            mult_above_axis = [1]
            mult_below_axis = [1]
            for mode in range(1, int(Nm / 2) + 1):
                cos = np.cos( mode * theta )
                sin = np.sin( mode * theta )
                mult_above_axis += [cos, sin]
                mult_below_axis += [ (-1) ** mode * cos, (-1) ** mode * sin ]
            mult_above_axis = np.array( mult_above_axis )
            mult_below_axis = np.array( mult_below_axis )
            # - Sum the modes
            F = get_data( series, component )  # (Extracts all modes)
            if coord_order is RZorder.mrz:
                F_total[Nr:, :] = np.tensordot( mult_above_axis,
                                                F, axes=(0, 0) )[:, :]
                F_total[:Nr, :] = np.tensordot( mult_below_axis,
                                                F, axes=(0, 0) )[::-1, :]
            elif coord_order is RZorder.mzr:
                F_total[:, Nr:] = np.tensordot( mult_above_axis,
                                                F, axes=(0, 0) )[:, :]
                F_total[:, :Nr] = np.tensordot( mult_below_axis,
                                                F, axes=(0, 0) )[:, ::-1]
            else:
                raise Exception(order_error_msg)
        elif m == 0:
            # Extract mode 0
            F = get_data( series, component, 0, 0 )
            if coord_order is RZorder.mrz:
                F_total[Nr:, :] = F[:, :]
                F_total[:Nr, :] = F[::-1, :]
            elif coord_order is RZorder.mzr:
                F_total[:, Nr:] = F[:, :]
                F_total[:, :Nr] = F[:, ::-1]
            else:
                raise Exception(order_error_msg)
        else:
            # Extract higher mode
            cos = np.cos( m * theta )
            sin = np.sin( m * theta )
            F_cos = get_data( series, component, 2 * m - 1, 0 )
            F_sin = get_data( series, component, 2 * m, 0 )
            F = cos * F_cos + sin * F_sin
            if coord_order is RZorder.mrz:
                F_total[Nr:, :] = F[:, :]
                F_total[:Nr, :] = (-1) ** m * F[::-1, :]
            elif coord_order is RZorder.mzr:
                F_total[:, Nr:] = F[:, :]
                F_total[:, :Nr] = (-1) ** m * F[:, ::-1]
            else:
                raise Exception(order_error_msg)

    # Perform slicing if needed
    if slice_across is not None:
        # Slice field and clear metadata
        inverted_axes_dict = {info.axes[key]: key for key in info.axes.keys()}
        for count, slice_across_item in enumerate(slice_across):
            slicing_index = inverted_axes_dict[slice_across_item]
            coord_array = getattr( info, slice_across_item )
            # Number of cells along the slicing direction
            n_cells = len(coord_array)
            # Index of the slice (prevent stepping out of the array)
            i_cell = int( 0.5 * (slice_relative_position[count] + 1.) * n_cells )
            i_cell = max( i_cell, 0 )
            i_cell = min( i_cell, n_cells - 1)
            F_total = np.take( F_total, [i_cell], axis=slicing_index )
        F_total = np.squeeze(F_total)
        # Remove the sliced labels from the FieldMetaInformation
        for slice_across_item in slice_across:
            info._remove_axis(slice_across_item)

    return F_total, info


# FIXME this looks like it can be generalized from already read meta-data
def get_grid_parameters( series, iteration, avail_fields, metadata ):
    """
    Return the parameters of the spatial grid (grid size and grid range)
    in two dictionaries

    Parameters:
    -----------
    series: openpmd_api.Series
        An open, readable openPMD-api series object

    iteration: integer
        Iteration from which parameters should be extracted

    avail_fields: list
       A list of the available fields
       e.g. ['B', 'E', 'rho']

    metadata: dictionary
      A dictionary whose keys are the fields of `avail_fields` and
      whose values are dictionaries that contain metadata (e.g. geometry)

    Returns:
    --------
    A tuple with `grid_size_dict` and `grid_range_dict`
    Both objects are dictionaries, with their keys being the labels of the axis
    of the grid (e.g. 'x', 'y', 'z')
    The values of `grid_size_dict` are the number of gridpoints along each axis
    The values of `grid_range_dict` are lists of two floats, which correspond
    to the min and max of the grid, along each axis.
    """
    it = series.iterations[iteration]

    # Pick field with the highest dimensionality ('3d'>'thetaMode'>'2d')
    # (This function is for the purpose of histogramming the particles;
    # in this case, the highest dimensionality ensures that more particle
    # quantities can be properly histogrammed.)
    geometry_ranking = {'1dcartesian': 0, '2dcartesian': 1,
                        'thetaMode': 2, '3dcartesian': 3}
    fields_ranking = [ geometry_ranking[ metadata[field]['geometry'] ]
                        for field in avail_fields ]
    index_best_field = fields_ranking.index( max(fields_ranking) )
    field_name = avail_fields[ index_best_field ]

    # Extract the dataset and and corresponding group
    # For field vector, extract the first component, to get the dataset
    field = it.meshes[field_name]
    component = next(field.items())[1]

    # Extract relevant quantities
    #   FIXME here and in h5py reader, we need to invert the order on 'F' for
    #         grid spacing/offset/position
    labels = field.axis_labels
    grid_spacing = np.array(field.grid_spacing) * field.grid_unit_SI
    grid_offset = np.array(field.grid_global_offset) * field.grid_unit_SI
    grid_size = component.shape
    if metadata[field_name]['geometry'] == 'thetaMode':
        # In thetaMode: skip the first number of dset.shape, as this
        # corresponds to the number of modes
        grid_size = component.shape[1:]

    # Build the dictionaries grid_size_dict and grid_range_dict
    grid_size_dict = {}
    grid_range_dict = {}
    for i in range(len(labels)):
        coord = labels[i]
        grid_size_dict[coord] = grid_size[i]
        grid_range_dict[coord] = \
            [ grid_offset[i], grid_offset[i] + grid_size[i] * grid_spacing[i] ]

    return grid_size_dict, grid_range_dict



"""
This test file is part of the openPMD-viewer.

It makes sure that the tutorial notebooks run without error.

Usage:
This file is meant to be run from the root directory of openPMD-viewer,
by any of the following commands
$ python tests/test_tutorials.py
$ py.test
$ python setup.py test

Copyright 2015-2016, openPMD-viewer contributors
Authors: Remi Lehe, Axel Huebl
License: 3-Clause-BSD-LBNL
"""

import os
import re


def test_tutorials():
    """Test all the tutorial notebooks"""

    # Go to the relative path where all tutorial notebooks are
    os.chdir('docs/source/tutorials')
    tutorial_notebooks = [filename for filename in os.listdir('./')
                          if filename[-6:] == '.ipynb']

    # Loop through the tutorials and test them
    for notebook_name in tutorial_notebooks:

        # Do a first pass where only the non-IPython features are tested.
        # (This gives better debugging information.)
        # The notebook is converted to a standard Python script and
        # run directly with `execfile`
        script_name = notebook_name[:-6] + '.py'
        os.system('jupyter nbconvert --to=python %s' % notebook_name)
        clean_ipython_features(script_name)
        try:
            response = os.system('python3 ' + script_name)
            assert response == 0
        except:
            # now we might want to know the script that was executed
            print(open(script_name).read())
            # re-raise same exception to make test fail
            raise
        os.remove(script_name)


def clean_ipython_features(script_name):
    """
    Rewrites the Python script `script_name` by removing
    all the IPython-specific commands

    Parameters
    ----------
    script_name: string
        Name of the Python script
    """
    # Read the script file
    with open(script_name) as script_file:
        lines = script_file.readlines()

    # Go through the lines and replace the IPython-specific commands
    # using regular expressions
    for i in range(len(lines)):

        # Replace the lines that activate matplotlib in a notebook
        # by a line that selects the PS backend
        if re.search(r"get_ipython.*matplotlib", lines[i]) is not None:
            lines[i] = "import matplotlib; matplotlib.use('ps')\n"

        # Discard the lines that use in-notebook documentation
        if re.search(r"get_ipython.*pinfo", lines[i]) is not None:
            lines[i] = ''

        # Discard the lines that use the GUI
        if re.match(r"[\w]*\.slider", lines[i]) is not None:
            lines[i] = ''

        # Replace the lines that call the OS by proper lines
        if re.match(r"[ ]*get_ipython\(\)\.system", lines[i]) is not None:
            matched = re.match(r"([ ]*)get_ipython\(\)\.system(.*)", lines[i])
            spaces = matched.groups()[0]
            command_line = matched.groups()[1]
            lines[i] = '%simport os; os.system%s\n' % (spaces, command_line)

    # Write the cleaned file
    with open(script_name, 'w') as script_file:
        for line in lines:
            script_file.write(line)


if __name__ == '__main__':
    test_tutorials()



"""
This file is part of the openPMD-viewer.

It defines a function that reads a species record component (data & meta)
from an openPMD file

Copyright 2020, openPMD-viewer contributors
Authors: Axel Huebl
License: 3-Clause-BSD-LBNL
"""

import numpy as np
from scipy import constants
from .utilities import get_data


def read_species_data(series, iteration, species_name, component_name,
                      extensions):
    """
    Extract a given species' record_comp

    Parameters
    ----------
    series: openpmd_api.Series
        An open, readable openPMD-api series object

    iteration: integer
        Iteration from which parameters should be extracted

    species_name: string
        The name of the species to extract (in the openPMD file)

    component_name: string
        The record component to extract
        Either 'x', 'y', 'z', 'ux', 'uy', 'uz', or 'w'

    extensions: list of strings
        The extensions that the current OpenPMDTimeSeries complies with
    """
    it = series.iterations[iteration]

    # Translate the record component to the openPMD format
    dict_record_comp = {'x': ['position', 'x'],
                        'y': ['position', 'y'],
                        'z': ['position', 'z'],
                        'ux': ['momentum', 'x'],
                        'uy': ['momentum', 'y'],
                        'uz': ['momentum', 'z'],
                        'w': ['weighting', None]}
    
    if component_name in dict_record_comp:
        ompd_record_name, ompd_record_comp_name = \
            dict_record_comp[component_name]
    elif component_name.find('/') != -1:
        ompd_record_name, ompd_record_comp_name = \
            component_name.split('/')
    else:
        ompd_record_name = component_name
        ompd_record_comp_name = None

    # Extract the right dataset
    species = it.particles[species_name]
    record = species[ompd_record_name]
    if record.scalar:
        component = next(record.items())[1]
    else:
        component = record[ompd_record_comp_name]

    if ompd_record_name == 'id':
        output_type = np.uint64
    else:
        output_type = np.float64
    data = get_data( series, component, output_type=output_type )

    # For ED-PIC: if the data is weighted for a full macroparticle,
    # divide by the weight with the proper power
    # (Skip this if the current record component is the weight itself)
    if 'ED-PIC' in extensions and ompd_record_name != 'weighting':
        macro_weighted = record.get_attribute('macroWeighted')
        weighting_power = record.get_attribute('weightingPower')
        if (macro_weighted == 1) and (weighting_power != 0):
            w_component = next(species['weighting'].items())[1]
            w = get_data( w_component )
            data *= w ** (-weighting_power)

    # - Return positions, with an offset
    if component_name in ['x', 'y', 'z']:
        offset = get_data(series, species['positionOffset'][component_name])
        data += offset
    # - Return momentum in normalized units
    elif component_name in ['ux', 'uy', 'uz' ]:
        mass_component = next(species['mass'].items())[1]
        m = get_data(series, mass_component)
        # Normalize only if the particle mass is non-zero
        if np.all( m != 0 ):
            norm_factor = 1. / (m * constants.c)
            data *= norm_factor


    # Return the data
    return data



"""
This file is part of the openPMD-viewer.

It defines functions that can read the fields from an HDF5 file.

Copyright 2015-2016, openPMD-viewer contributors
Author: Remi Lehe
License: 3-Clause-BSD-LBNL
"""

import h5py
import numpy as np
from .utilities import get_shape, get_data, join_infile_path
from ...data_order import RZorder, order_error_msg
from openpmd_viewer.openpmd_timeseries.field_metainfo import FieldMetaInformation
from openpmd_viewer.openpmd_timeseries.utilities import construct_3d_from_circ


def read_field_cartesian( filename, iteration, field, coord, axis_labels,
                          slice_relative_position, slice_across ):
    """
    Extract a given field from an HDF5 file in the openPMD format,
    when the geometry is cartesian (1d, 2d or 3d).

    Parameters
    ----------
    filename : string
       The absolute path to the HDF5 file

    iteration : int
        The iteration at which to obtain the data

    field : string, optional
       Which field to extract

    coord : string, optional
       Which component of the field to extract

    axis_labels: list of strings
       The name of the dimensions of the array (e.g. ['x', 'y', 'z'])

    slice_across : list of str or None
       Direction(s) across which the data should be sliced
       Elements can be:
         - 1d: 'z'
         - 2d: 'x' and/or 'z'
         - 3d: 'x' and/or 'y' and/or 'z'
       Returned array is reduced by 1 dimension per slicing.

    slice_relative_position : list of float or None
       Number(s) between -1 and 1 that indicate where to slice the data,
       along the directions in `slice_across`
       -1 : lower edge of the simulation box
       0 : middle of the simulation box
       1 : upper edge of the simulation box

    Returns
    -------
    A tuple with
       F : a ndarray containing the required field
       info : a FieldMetaInformation object
       (contains information about the grid; see the corresponding docstring)
    """
    # Open the HDF5 file
    dfile = h5py.File( filename, 'r' )
    # Extract the iteration
    it = dfile["/data/" + str(iteration)]

    # Extract the dataset and corresponding group
    if coord is None:
        field_path = field
    else:
        field_path = join_infile_path( field, coord )
    group, dset = find_dataset( dfile, iteration, field_path )

    # Dimensions of the grid
    shape = list( get_shape( dset ) )
    grid_spacing = list( group.attrs['gridSpacing'] )
    global_offset = list( group.attrs['gridGlobalOffset'] )

    # Current simulation time
    time = (it.attrs['time'] + group.attrs['timeOffset']) * it.attrs['timeUnitSI']

    # Slice selection
    if slice_across is not None:
        # Get the integer that correspond to the slicing direction
        list_slicing_index = []
        list_i_cell = []
        for count, slice_across_item in enumerate(slice_across):
            slicing_index = axis_labels.index(slice_across_item)
            list_slicing_index.append(slicing_index)
            # Number of cells along the slicing direction
            n_cells = shape[ slicing_index ]
            # Index of the slice (prevent stepping out of the array)
            i_cell = int( 0.5 * (slice_relative_position[count] + 1.) * n_cells )
            i_cell = max( i_cell, 0 )
            i_cell = min( i_cell, n_cells - 1)
            list_i_cell.append(i_cell)

        # Remove metainformation relative to the slicing index
        # Successive pops starting from last coordinate to slice
        shape = [ x for index, x in enumerate(shape)
                  if index not in list_slicing_index ]
        grid_spacing = [ x for index, x in enumerate(grid_spacing)
                         if index not in list_slicing_index ]
        global_offset = [ x for index, x in enumerate(global_offset)
                          if index not in list_slicing_index ]
        axis_labels = [ x for index, x in enumerate(axis_labels)
                         if index not in list_slicing_index ]

        axes = { i: axis_labels[i] for i in range(len(axis_labels)) }
        # Extract data
        F = get_data( dset, list_i_cell, list_slicing_index )
        info = FieldMetaInformation( axes, shape, grid_spacing, global_offset,
                group.attrs['gridUnitSI'], dset.attrs['position'],
                time, iteration )
    else:
        F = get_data( dset )
        axes = { i: axis_labels[i] for i in range(len(axis_labels)) }
        info = FieldMetaInformation( axes, F.shape,
            group.attrs['gridSpacing'], group.attrs['gridGlobalOffset'],
            group.attrs['gridUnitSI'], dset.attrs['position'],
            time, iteration )

    # Close the file
    dfile.close()
    return( F, info )


def read_field_circ( filename, iteration, field, coord,
                     slice_relative_position, slice_across, m=0, theta=0.,
                     max_resolution_3d=None ):
    """
    Extract a given field from an HDF5 file in the openPMD format,
    when the geometry is thetaMode

    Parameters
    ----------
    filename : string
       The absolute path to the HDF5 file

    iteration : int
        The iteration at which to obtain the data

    field : string, optional
       Which field to extract

    coord : string, optional
       Which component of the field to extract

    m : int or string, optional
       The azimuthal mode to be extracted

    theta : float or None
       Angle of the plane of observation with respect to the x axis
       If `theta` is not None, then this function returns a 2D array
       corresponding to the plane of observation given by `theta` ;
       otherwise it returns a full 3D Cartesian array

    slice_across : list of str or None
       Direction(s) across which the data should be sliced
       Elements can be 'r' and/or 'z'
       Returned array is reduced by 1 dimension per slicing.

    slice_relative_position : list of float or None
       Number(s) between -1 and 1 that indicate where to slice the data,
       along the directions in `slice_across`
       -1 : lower edge of the simulation box
       0 : middle of the simulation box
       1 : upper edge of the simulation box

    max_resolution_3d : list of int or None
        Maximum resolution that the 3D reconstruction of the field (when
        `theta` is None) can have. The list should contain two values,
        e.g. `[200, 100]`, indicating the maximum longitudinal and transverse
        resolution, respectively. This is useful for performance reasons,
        particularly for 3D visualization.

    Returns
    -------
    A tuple with
       F : a 3darray or 2darray containing the required field,
           depending on whether `theta` is None or not
       info : a FieldMetaInformation object
       (contains information about the grid; see the corresponding docstring)
    """
    # Open the HDF5 file
    dfile = h5py.File( filename, 'r' )
    # Extract the iteration
    it = dfile["/data/" + str(iteration)]

    # Extract the dataset and corresponding group
    if coord is None:
        field_path = field
    else:
        field_path = join_infile_path( field, coord )
    group, dset = find_dataset( dfile, iteration, field_path )

    # Extract the metainformation
    coord_labels = {ii: coord.decode() for (ii,coord) in
                                enumerate(group.attrs['axisLabels'])}
    coord_label_str = ''.join(coord_labels.values())
    coord_label_str = 'm' + coord_label_str
    coord_order = RZorder[coord_label_str]

    # Current simulation time
    time = (it.attrs['time'] + group.attrs['timeOffset']) * it.attrs['timeUnitSI']

    if coord_order == RZorder.mrz:
        Nm, Nr, Nz = get_shape( dset )
        N_pair = (Nr, Nz)
    elif coord_order == RZorder.mzr:
        Nm, Nz, Nr = get_shape( dset )
        N_pair = (Nz, Nr)
    else:
        raise Exception(order_error_msg)
    info = FieldMetaInformation( coord_labels, N_pair,
        group.attrs['gridSpacing'], group.attrs['gridGlobalOffset'],
        group.attrs['gridUnitSI'], dset.attrs['position'], time,
        iteration, thetaMode=True )

    # Convert to a 3D Cartesian array if theta is None
    if theta is None:

        # Get cylindrical info
        rmax = info.rmax
        inv_dr = 1./info.dr
        Fcirc = get_data( dset )  # (Extracts all modes)
        if m == 'all':
            modes = [ mode for mode in range(0, int(Nm / 2) + 1) ]
        else:
            modes = [ m ]
        modes = np.array( modes, dtype='int' )
        nmodes = len(modes)

        # If necessary, reduce resolution of 3D reconstruction
        if max_resolution_3d is not None:
            max_res_lon, max_res_transv = max_resolution_3d
            if Nz > max_res_lon:
                # Calculate excess of elements along z
                excess_z = int(np.round(Nz/max_res_lon))
                # Preserve only one every excess_z elements
                if coord_order is RZorder.mrz:
                    Fcirc = Fcirc[:, :, ::excess_z]
                elif coord_order is RZorder.mzr:
                    Fcirc = Fcirc[:, ::excess_z, :]
                else:
                    raise Exception(order_error_msg)

                # Update info accordingly
                info.z = info.z[::excess_z]
                info.dz = info.z[1] - info.z[0]
            if Nr > max_res_transv/2:
                # Calculate excess of elements along r
                excess_r = int(np.round(Nr/(max_res_transv/2)))
                # Preserve only one every excess_r elements
                if coord_order is RZorder.mrz:
                    Fcirc = Fcirc[:, ::excess_r, :]
                elif coord_order is RZorder.mzr:
                    Fcirc = Fcirc[:, :, ::excess_r]
                else:
                    raise Exception(order_error_msg)

                # Update info and necessary parameters accordingly
                info.r = info.r[::excess_r]
                info.dr = info.r[1] - info.r[0]
                inv_dr = 1./info.dr
                # Update Nr after reducing radial resolution.
                if coord_order is RZorder.mrz:
                    Nr = Fcirc.shape[1]
                elif coord_order is RZorder.mzr:
                    Nr = Fcirc.shape[2]
                else:
                    raise Exception(order_error_msg)


        # Convert cylindrical data to Cartesian data
        info._convert_cylindrical_to_3Dcartesian()
        nx, ny, nz = len(info.x), len(info.y), len(info.z)
        F_total = np.zeros( (nx, ny, nz) )
        construct_3d_from_circ( F_total, Fcirc, info.x, info.y, modes,
            nx, ny, nz, Nr, nmodes, inv_dr, rmax, coord_order )

    else:

        # Extract the modes and recombine them properly
        if coord_order is RZorder.mrz:
            F_total = np.zeros( (2 * Nr, Nz ) )
        elif coord_order is RZorder.mzr:
            F_total = np.zeros( (Nz, 2 * Nr ) )
        else:
            raise Exception(order_error_msg)
        if m == 'all':
            # Sum of all the modes
            # - Prepare the multiplier arrays
            mult_above_axis = [1]
            mult_below_axis = [1]
            for mode in range(1, int(Nm / 2) + 1):
                cos = np.cos( mode * theta )
                sin = np.sin( mode * theta )
                mult_above_axis += [cos, sin]
                mult_below_axis += [ (-1) ** mode * cos, (-1) ** mode * sin ]
            mult_above_axis = np.array( mult_above_axis )
            mult_below_axis = np.array( mult_below_axis )
            # - Sum the modes
            F = get_data( dset )  # (Extracts all modes)
            if coord_order is RZorder.mrz:
                F_total[Nr:, :] = np.tensordot( mult_above_axis,
                                                F, axes=(0, 0) )[:, :]
                F_total[:Nr, :] = np.tensordot( mult_below_axis,
                                                F, axes=(0, 0) )[::-1, :]
            elif coord_order is RZorder.mzr:
                F_total[:, Nr:] = np.tensordot( mult_above_axis,
                                                F, axes=(0, 0) )[:, :]
                F_total[:, :Nr] = np.tensordot( mult_below_axis,
                                                F, axes=(0, 0) )[:, ::-1]
        elif m == 0:
            # Extract mode 0
            F = get_data( dset, 0, 0 )
            if coord_order is RZorder.mrz:
                F_total[Nr:, :] = F[:, :]
                F_total[:Nr, :] = F[::-1, :]
            elif coord_order is RZorder.mzr:
                F_total[:, Nr:] = F[:, :]
                F_total[:, :Nr] = F[:, ::-1]
            else:
                raise Exception(order_error_msg)

        else:
            # Extract higher mode
            cos = np.cos( m * theta )
            sin = np.sin( m * theta )
            F_cos = get_data( dset, 2 * m - 1, 0 )
            F_sin = get_data( dset, 2 * m, 0 )
            F = cos * F_cos + sin * F_sin
            if coord_order is RZorder.mrz:
                F_total[Nr:, :] = F[:, :]
                F_total[:Nr, :] = (-1) ** m * F[::-1, :]
            elif coord_order is RZorder.mzr:
                F_total[:, Nr:] = F[:, :]
                F_total[:, :Nr] = (-1) ** m * F[:, ::-1]
            else:
                raise Exception(order_error_msg)

    # Perform slicing if needed
    if slice_across is not None:
        # Slice field and clear metadata
        inverted_axes_dict = {info.axes[key]: key for key in info.axes.keys()}
        for count, slice_across_item in enumerate(slice_across):
            slicing_index = inverted_axes_dict[slice_across_item]
            coord_array = getattr( info, slice_across_item )
            # Number of cells along the slicing direction
            n_cells = len(coord_array)
            # Index of the slice (prevent stepping out of the array)
            i_cell = int( 0.5 * (slice_relative_position[count] + 1.) * n_cells )
            i_cell = max( i_cell, 0 )
            i_cell = min( i_cell, n_cells - 1)
            F_total = np.take( F_total, [i_cell], axis=slicing_index )
        F_total = np.squeeze(F_total)
        # Remove the sliced labels from the FieldMetaInformation
        for slice_across_item in slice_across:
            info._remove_axis(slice_across_item)

    # Close the file
    dfile.close()

    return( F_total, info )


def find_dataset( dfile, iteration, field_path ):
    """
    Extract the dataset that corresponds to field_path,
    and the corresponding group

    (In the case of scalar records, the group and the dataset are identical.
    In the case of vector records, the group contains all the components
    and the dataset corresponds to one given component.)

    Parameters
    ----------
    dfile: an h5Py.File object
       The file from which to extract the dataset

    field_path : string
       The relative path to the requested field, from the openPMD meshes path
       (e.g. 'rho', 'E/r', 'B/x')

    Returns
    -------
    A tuple with:
    - an h5py.Group object
    - an h5py.Dataset object
    """
    # Find the meshes path
    base_path = '/data/{0}'.format( iteration )
    relative_meshes_path = dfile.attrs["meshesPath"].decode()

    # Get the proper dataset
    full_field_path = join_infile_path(
        base_path, relative_meshes_path, field_path )
    dset = dfile[ full_field_path ]
    # Get the proper group
    group_path = field_path.split('/')[0]
    full_group_path = join_infile_path(
        base_path, relative_meshes_path, group_path )
    group = dfile[ full_group_path ]

    return( group, dset )


def get_grid_parameters( filename, iteration, avail_fields, metadata ):
    """
    Return the parameters of the spatial grid (grid size and grid range)
    in two dictionaries

    Parameters:
    -----------
    filename : string
       The absolute path to the HDF5 file

    iteration : int
        The iteration at which to obtain the data

    avail_fields: list
       A list of the available fields
       e.g. ['B', 'E', 'rho']

    metadata: dictionary
      A dictionary whose keys are the fields of `avail_fields` and
      whose values are dictionaries that contain metadata (e.g. geometry)

    Returns:
    --------
    A tuple with `grid_size_dict` and `grid_range_dict`
    Both objects are dictionaries, with their keys being the labels of the axis
    of the grid (e.g. 'x', 'y', 'z')
    The values of `grid_size_dict` are the number of gridpoints along each axis
    The values of `grid_range_dict` are lists of two floats, which correspond
    to the min and max of the grid, along each axis.
    """
    # Open the HDF5 file
    dfile = h5py.File( filename, 'r' )
    # Pick field with the highest dimensionality ('3d'>'thetaMode'>'2d')
    # (This function is for the purpose of histogramming the particles;
    # in this case, the highest dimensionality ensures that more particle
    # quantities can be properly histogrammed.)
    geometry_ranking = {'1dcartesian': 0, '2dcartesian': 1,
                        'thetaMode': 2, '3dcartesian': 3}
    fields_ranking = [ geometry_ranking[ metadata[field]['geometry'] ]
                        for field in avail_fields ]
    index_best_field = fields_ranking.index( max(fields_ranking) )
    field_name = avail_fields[ index_best_field ]

    # Get the corresponding field data
    group, dset = find_dataset( dfile, iteration, field_name )
    if metadata[field_name]['type'] == 'vector':
        # For field vector, extract the first coordinate, to get the dataset
        first_coord = next(iter(group.keys()))
        dset = group[first_coord]

    # Extract relevant quantities
    labels = group.attrs['axisLabels']
    grid_spacing = group.attrs['gridSpacing'] * group.attrs['gridUnitSI']
    grid_offset = group.attrs['gridGlobalOffset'] * group.attrs['gridUnitSI']
    grid_size = dset.shape
    if metadata[field_name]['geometry'] == 'thetaMode':
        # In thetaMode: skip the first number of dset.shape, as this
        # corresponds to the number of modes
        grid_size = dset.shape[1:]

    # Build the dictionaries grid_size_dict and grid_range_dict
    grid_size_dict = {}
    grid_range_dict = {}
    for i in range(len(labels)):
        coord = labels[i].decode()
        grid_size_dict[coord] = grid_size[i]
        grid_range_dict[coord] = \
            [ grid_offset[i], grid_offset[i] + grid_size[i] * grid_spacing[i] ]
    # Close the file
    dfile.close()
    return( grid_size_dict, grid_range_dict )



"""
This file is part of the openPMD-viewer.

It defines the ParticleTracker class.

Copyright 2015-2016, openPMD-viewer contributors
Authors: Remi Lehe
License: 3-Clause-BSD-LBNL
"""
import numpy as np
from .numba_wrapper import jit

class ParticleTracker( object ):
    """
    Class that allows to select particles at a given iteration
    (by initializing an instance of this class) and then
    to return the same particles at another iteration (by passing
    this instance as the argument `select` of the method `get_particle`
    of an `OpenPMDTimeSeries`)

    Usage
    -----
    Here is a minimal example of how this class is used.
    In this example, all the particles in the simulation box at iteration 300
    are selected, and then the position of the same particles at iteration 400
    (or at least of those particles that remained in the simulation box at
    iteration 400) are returned.
    ```
    >>> ts = OpenPMDTimeSeries('./hdf5_directory/')
    >>> pt = ParticleTracker( ts, iteration=300 )
    >>> x, = ts.get_particle( ['x'], select=pt, iteration=400 )
    ```
    For more details on the API of ParticleTracker, see the docstring of
    the `__init__` method of this class.

    Note
    ----
    `ParticleTracker` requires the `id` of the particles
    to be stored in the openPMD files.
    """

    def __init__(self, ts, species=None, t=None,
                iteration=None, select=None, preserve_particle_index=False):
        """
        Initialize an instance of `ParticleTracker`: select particles at
        a given iteration, so that they can be retrieved at a later iteration.

        Parameters
        ----------
        ts: an OpenPMDTimeSeries object
            Contains the data on the particles

        species: string
            A string indicating the name of the species
            This is optional if there is only one species

        t : float (in seconds), optional
            Time at which to obtain the data (if this does not correspond to
            an existing iteration, the closest existing iteration will be used)
            Either `t` or `iteration` should be given by the user.

        iteration : int
            The iteration at which to obtain the data
            Either `t` or `iteration` should be given by the user.

        select: dict or 1darray of int, optional
            Either None or a dictionary of rules
            to select the particles, of the form
            'x' : [-4., 10.]  (Particles having x between -4 and 10)
            'ux' : [-0.1, 0.1] (Particles having ux between -0.1 and 0.1 mc)
            'uz' : [5., None]  (Particles with uz above 5 mc).
            Can also be a 1d array of interegers corresponding to the
            selected particles `id`

        preserve_particle_index: bool, optional
            When retrieving particles at a several iterations,
            (for instance, with:
            ```
            >>> x1, = ts.get_particle( ['x'], select=pt, iteration=400 )
            >>> x2, = ts.get_particle( ['x'], select=pt, iteration=500 )
            ```
            it is sometimes important that the same individual particle has
            the same index in the array `x1` and `x2`.
            Using `preserve_particle_index=True` ensures that this is the case.
            However, this means that, for a particle that becomes absent at
            a later iteration, its index in the array has to be filled also.
            In this case, a NaN is returned at the index of this particle.
            When `preserve_particle_index=False`, no NaN is returned (the
            returned array is simply smaller when particles are absent) but
            then it is not garanteed that a given particle keeps the same index
        """

        # Extract or load the particle id and sort them
        if (type(select) is dict) or (select is None):
            self.selected_pid, = ts.get_particle(['id'], species=species,
                                    select=select, t=t, iteration=iteration)
        elif (type(select) is np.ndarray):
            self.selected_pid = select

        self.selected_pid.sort()

        # Register a few metadata
        self.N_selected = len( self.selected_pid )
        self.species = species
        self.preserve_particle_index = preserve_particle_index


    def extract_tracked_particles( self, iteration, data_reader, data_list,
                                    species, extensions ):
        """
        Select the elements of each particle quantities in data_list,
        so as to only return those that correspond to the tracked particles

        Parameters
        ----------
        iteration: int
            The iteration at which to extract the particles

        data_reader: a DataReader object
            Used in order to extract the macroparticle IDs

        data_list: list of 1darrays
            A list of arrays with one element per macroparticle, that represent
            different particle quantities

        species: string
            Name of the species being requested

        extensions: list of strings
            The extensions that the current OpenPMDTimeSeries complies with

        Returns
        -------
        A list of 1darrays that correspond to data_list, but where only the
        particles that are tracked are kept. (NaNs may or may not be returned
        depending on whether `preserve_particle_index` was chosen at
        initialization)
        """
        # Extract the particle id, and get the extraction indices
        pid = data_reader.read_species_data(iteration, species, 'id', extensions)
        selected_indices = self.get_extraction_indices( pid )

        # For each particle quantity, select only the tracked particles
        for i in range(len(data_list)):
            if len(data_list[i]) > 1:  # Do not apply selection on scalars
                data_list[i] = self.extract_quantity(
                    data_list[i], selected_indices )

        return( data_list )

    def extract_quantity( self, q, selected_indices ):
        """
        Select the elements of the array `q`, so as to only return those
        that correspond to the tracked particles.

        Parameters
        ----------
        q: 1d array of floats or ints
            A particle quantity (one element per particle)

        selected_indices: 1d array of ints
            The indices (in array q) of the particles to be selected.
            If `preserve_particle_index` was selected to be True, this array
            contains -1 at the position of particles that are no longer present

        Returns
        -------
        selected_q: 1d array of floats or ints
            A particle quantity (one element per particles)
            where only the tracked particles are kept
        """
        # Extract the selected elements
        selected_q = q[ selected_indices ]

        # Handle the absent particles
        if self.preserve_particle_index:
            if q.dtype in [ np.float64, np.float32 ]:
                # Fill the position of absent particles by NaNs
                selected_q = np.where( selected_indices == -1,
                                        np.nan, selected_q)
            else:
                # The only non-float quantity in openPMD-viewer is particle id
                selected_q = self.selected_pid

        return( selected_q )

    def get_extraction_indices( self, pid ):
        """
        For each tracked particle (i.e. for each element of self.selected_pid)
        find the index of the same particle in the array `pid`

        Return these indices in an array, so that it can then be used to
        extract quantities (position, momentum, etc.) for the tracked particles

        Parameters
        ----------
        pid: 1darray of ints
            The id of each particle (one element per particle)

        Returns
        -------
        selected_indices: 1d array of ints
            The index of the tracked particles in the array `pid`
            If `preserve_particle_index` was selected to be True, this array
            contains -1 at the position of particles that are no longer present

        Note on the implementation
        --------------------------
        This could be implemented in brute force (i.e. for each element of
        `self.selected_pid`, search the entire array `pid` for the same
        element), but would be very costly.
        Instead, we sort the array `pid` (and keep track of the original
        pid, so as to be able to retrieve the indices) and use the fact
        that it is sorted in order to rapidly extract the elements
        that are also in `self.selected_pid` (which is also sorted)
        """
        # Sort the pid, and keep track of the original index
        # at which each pid was
        original_indices = pid.argsort().astype(np.int64)
        sorted_pid = pid[ original_indices ]

        # Extract only the indices for which sorted_pid is one of pid
        # in self.sselected_pid (i.e. which correpond to one
        # of the original particles)
        selected_indices = np.empty( self.N_selected, dtype=np.int64 )
        N_extracted = extract_indices(
            original_indices, selected_indices, sorted_pid,
            self.selected_pid, self.preserve_particle_index )

        # If there are less particles then self.N_selected
        # (i.e. not all the pid in self.selected_pid were in sorted_pid)
        if N_extracted < self.N_selected:
            if self.preserve_particle_index:
                # Finish filling the array with absent particles
                selected_indices[N_extracted:] = -1
            else:
                # Resize the array
                selected_indices = \
                    selected_indices[:N_extracted]

        return( selected_indices )

@jit
def extract_indices( original_indices, selected_indices,
                        pid, selected_pid, preserve_particle_index ):
    """
    Go through the sorted arrays `pid` and `selected_pid`, and record
    the indices (of the array `pid`) where they match, by storing them
    in the array `selected_indices` (this array is thus modified in-place)

    Return the number of elements that were filled in `selected_indices`
    """
    i = 0
    i_select = 0
    i_fill = 0
    N = len(pid)
    N_selected = len(selected_pid)

    # Go through both sorted arrays (pid and selected_pid) and match them.
    # i.e. whenever the same number appears in both arrays,
    # record the corresponding original index in selected_indices
    while i < N and i_select < N_selected:

        if pid[i] < selected_pid[i_select]:
            i += 1
        elif pid[i] == selected_pid[i_select]:
            selected_indices[i_fill] = original_indices[i]
            i_fill += 1
            i_select += 1
        elif pid[i] > selected_pid[i_select]:
            i_select += 1
            if preserve_particle_index:
                # Fill the index, to indicate that the particle is absent
                selected_indices[i_fill] = -1
                i_fill += 1

    return( i_fill )



"""
This file is part of the openPMD-viewer.

It defines a function that reads a species record component (data & meta)
from an openPMD file

Copyright 2015-2016, openPMD-viewer contributors
Authors: Remi Lehe, Axel Huebl
License: 3-Clause-BSD-LBNL
"""

import h5py
import numpy as np
from scipy import constants
from .utilities import get_data, join_infile_path


def read_species_data(filename, iteration, species, record_comp, extensions):
    """
    Extract a given species' record_comp

    Parameters
    ----------
    filename : string
       The absolute path to the HDF5 file

    iteration : int
        The iteration at which to obtain the data

    species: string
        The name of the species to extract (in the openPMD file)

    record_comp: string
        The record component to extract
        Either 'x', 'y', 'z', 'ux', 'uy', 'uz', or 'w'

    extensions: list of strings
        The extensions that the current OpenPMDTimeSeries complies with
    """
    # Open the HDF5 file
    dfile = h5py.File( filename, 'r' )
    # Translate the record component to the openPMD format
    dict_record_comp = {'x': 'position/x',
                        'y': 'position/y',
                        'z': 'position/z',
                        'ux': 'momentum/x',
                        'uy': 'momentum/y',
                        'uz': 'momentum/z',
                        'w': 'weighting'}
    if record_comp in dict_record_comp:
        opmd_record_comp = dict_record_comp[record_comp]
    else:
        opmd_record_comp = record_comp

    # Open the HDF5 file
    base_path = '/data/{0}'.format( iteration )
    particles_path = dfile.attrs['particlesPath'].decode()

    # Extract the right dataset
    species_grp = dfile[
        join_infile_path(base_path, particles_path, species) ]
    if opmd_record_comp == 'id':
        output_type = np.uint64
    else:
        output_type = np.float64
    data = get_data( species_grp[ opmd_record_comp ], output_type=output_type )

    # For ED-PIC: if the data is weighted for a full macroparticle,
    # divide by the weight with the proper power
    # (Skip this if the current record component is the weight itself)
    if 'ED-PIC' in extensions and opmd_record_comp != 'weighting':
        opmd_record = opmd_record_comp.split('/')[0]
        record_dset = species_grp[ opmd_record ]
        macro_weighted = record_dset.attrs['macroWeighted']
        weighting_power = record_dset.attrs['weightingPower']
        if (macro_weighted == 1) and (weighting_power != 0):
            w = get_data( species_grp[ 'weighting' ] )
            data *= w ** (-weighting_power)

    # - Return positions, with an offset
    if record_comp in ['x', 'y', 'z']:
        offset = get_data(species_grp['positionOffset/%s' % record_comp])
        data += offset
    # - Return momentum in normalized units
    elif record_comp in ['ux', 'uy', 'uz' ]:
        m = get_data(species_grp['mass'])
        # Normalize only if the particle mass is non-zero
        if np.all( m != 0 ):
            norm_factor = 1. / (m * constants.c)
            data *= norm_factor

    # Close the file
    dfile.close()
    # Return the data
    return(data)



"""
This file is part of the openPMD-viewer

It routes the calls to the data reader to either
the h5py data reader, or to openpmd-api.

Copyright 2020, openPMD-viewer contributors
Authors: Remi Lehe
License: 3-Clause-BSD-LBNL
"""
import numpy as np
import os
import re

available_backends = []

try:
    import openpmd_api as io
    from . import io_reader
    available_backends.append('openpmd-api')
except ImportError:
    pass

try:
    from . import h5py_reader
    available_backends.append('h5py')
except ImportError:
    pass

if len(available_backends) == 0:
    raise ImportError('No openPMD backend found.\n'
        'Please install either `h5py` or `openpmd-api`:\n'
        'e.g. with `pip install h5py` or `pip install openpmd-api`')

class DataReader( object ):
    """
    Class that performs various type of access the openPMD file.

    The methods of this class are agnostic of the actual backend package
    used in order to access the openPMD file (e.g. h5py or openpmd-api).
    The backend that is used in practice depends on which package is
    available on the current environment.
    """

    def __init__(self, backend):
        """
        Initialize the DataReader class.
        """
        self.backend = backend

        # Point to the correct reader module
        if self.backend == 'h5py':
            self.iteration_to_file = {}
        elif self.backend == 'openpmd-api':
            pass
        else:
            raise RuntimeError('Unknown backend: %s' % self.backend)

    def list_iterations(self, path_to_dir):
        """
        Return a list of the iterations that correspond to the files
        in this directory. (The correspondance between iterations and
        files is stored internally.)

        Parameter
        ---------
        path_to_dir : string
            The path to the directory where the hdf5 files are.

        Returns
        -------
        an array of integers which correspond to the iteration of each file
        (in sorted order)
        """
        if self.backend == 'h5py':
            iterations, iteration_to_file = \
                h5py_reader.list_files( path_to_dir )
            # Store dictionary of correspondence between iteration and file
            self.iteration_to_file = iteration_to_file
            if len(iterations) == 0:
                raise RuntimeError(
                    "Found no valid files in directory {0}.\n"
                    "Please check that this is the path to the openPMD files."
                    "Valid files must have the extension '.h5' if you "
                    "use the `h5py` backend. For ADIOS '.bp' and other files, "
                    "please install the `openpmd-api` package."
                    .format(path_to_dir))
        elif self.backend == 'openpmd-api':
            # guess file ending from first file in directory
            first_file_name = None

            is_single_file = os.path.isfile(path_to_dir)
            if is_single_file:
                first_file_name = path_to_dir
            else:
                for file_name in os.listdir( path_to_dir ):
                    if file_name.split(os.extsep)[-1] in io.file_extensions:
                        first_file_name = file_name
            if first_file_name is None:
                raise RuntimeError(
                    "Found no valid files in directory {0}.\n"
                    "Please check that this is the path to the openPMD files."
                    "(valid files must have one of the following extensions: {1})"
                    .format(path_to_dir, io.file_extensions))

            if is_single_file:
                file_path = path_to_dir
                series_name = file_path
            else:
                # match last occurance of integers and replace with %T wildcards
                # examples: data00000100.h5 diag4_00000500.h5 io12.0.bp
                #           te42st.1234.yolo.json scan7_run14_data123.h5
                file_path = re.sub(r'(\d+)(\.(?!\d).+$)', r'%T\2', first_file_name)
                series_name = os.path.join( path_to_dir, file_path)

            self.series = io.Series(
                series_name,
                io.Access.read_only )
            iterations = np.array( self.series.iterations )

        return iterations

    def read_openPMD_params(self, iteration, extract_parameters=True):
        """
        Extract the time and some openPMD parameters from a file

        Parameter
        ---------
        iteration: int
            The iteration at which the parameters should be extracted

        extract_parameters: bool, optional
            Whether to extract all parameters or only the time
            (Function execution is faster when extract_parameters is False)

        Returns
        -------
        A tuple with:
        - A float corresponding to the time of this iteration in SI units
        - A dictionary containing several parameters, such as the geometry, etc
         When extract_parameters is False, the second argument returned is None
        """
        if self.backend == 'h5py':
            filename = self.iteration_to_file[iteration]
            return h5py_reader.read_openPMD_params(
                    filename, iteration, extract_parameters)

        elif self.backend == 'openpmd-api':
            return io_reader.read_openPMD_params(
                    self.series, iteration, extract_parameters)

    def read_field_cartesian( self, iteration, field, coord, axis_labels,
                          slice_relative_position, slice_across ):
        """
        Extract a given field from an openPMD file in the openPMD format,
        when the geometry is cartesian (1d, 2d or 3d).

        Parameters
        ----------
        iteration : int
           The iteration at which to extract the fields

        field : string, optional
           Which field to extract

        coord : string, optional
           Which component of the field to extract

        axis_labels: list of strings
           The name of the dimensions of the array (e.g. ['x', 'y', 'z'])

        slice_across : list of str or None
           Direction(s) across which the data should be sliced
           Elements can be:
             - 1d: 'z'
             - 2d: 'x' and/or 'z'
             - 3d: 'x' and/or 'y' and/or 'z'
           Returned array is reduced by 1 dimension per slicing.

        slice_relative_position : list of float or None
           Number(s) between -1 and 1 that indicate where to slice the data,
           along the directions in `slice_across`
           -1 : lower edge of the simulation box
           0 : middle of the simulation box
           1 : upper edge of the simulation box

        Returns
        -------
        A tuple with
           F : a ndarray containing the required field
           info : a FieldMetaInformation object
           (contains information about the grid; see the corresponding docstring)
        """
        if self.backend == 'h5py':
            filename = self.iteration_to_file[iteration]
            return h5py_reader.read_field_cartesian(
                filename, iteration, field, coord, axis_labels,
                slice_relative_position, slice_across )
        elif self.backend == 'openpmd-api':
            return io_reader.read_field_cartesian(
                self.series, iteration, field, coord, axis_labels,
                slice_relative_position, slice_across )

    def read_field_circ( self, iteration, field, coord, slice_relative_position,
                        slice_across, m=0, theta=0., max_resolution_3d=None ):
        """
        Extract a given field from an openPMD file in the openPMD format,
        when the geometry is thetaMode

        Parameters
        ----------
        iteration : int
           The iteration at which to extract the fields

        field : string, optional
           Which field to extract
           Either 'rho', 'E', 'B' or 'J'

        coord : string, optional
           Which component of the field to extract
           Either 'r', 't' or 'z'

        m : int or string, optional
           The azimuthal mode to be extracted

        theta : float or None
           Angle of the plane of observation with respect to the x axis
           If `theta` is not None, then this function returns a 2D array
           corresponding to the plane of observation given by `theta` ;
           otherwise it returns a full 3D Cartesian array

        slice_across : list of str or None
           Direction(s) across which the data should be sliced
           Elements can be 'r' and/or 'z'
           Returned array is reduced by 1 dimension per slicing.

        slice_relative_position : list of float or None
           Number(s) between -1 and 1 that indicate where to slice the data,
           along the directions in `slice_across`
           -1 : lower edge of the simulation box
           0 : middle of the simulation box
           1 : upper edge of the simulation box

        max_resolution_3d : list of int or None
            Maximum resolution that the 3D reconstruction of the field (when
            `theta` is None) can have. The list should contain two values,
            e.g. `[200, 100]`, indicating the maximum longitudinal and
            transverse resolution, respectively. This is useful for
            performance reasons, particularly for 3D visualization.

        Returns
        -------
        A tuple with
           F : a 3darray or 2darray containing the required field,
               depending on whether `theta` is None or not
           info : a FieldMetaInformation object
           (contains information about the grid; see the corresponding docstring)
        """
        if self.backend == 'h5py':
            filename = self.iteration_to_file[iteration]
            return h5py_reader.read_field_circ(
                filename, iteration, field, coord, slice_relative_position,
                slice_across, m, theta, max_resolution_3d )
        elif self.backend == 'openpmd-api':
            return io_reader.read_field_circ(
                self.series, iteration, field, coord, slice_relative_position,
                slice_across, m, theta, max_resolution_3d )

    def read_species_data( self, iteration, species, record_comp, extensions):
        """
        Extract a given species' record_comp

        Parameters
        ----------
        iteration: int
            The iteration at which to extract the species data

        species: string
            The name of the species to extract (in the openPMD file)

        record_comp: string
            The record component to extract
            Either 'x', 'y', 'z', 'ux', 'uy', 'uz', or 'w'

        extensions: list of strings
            The extensions that the current OpenPMDTimeSeries complies with
        """
        if self.backend == 'h5py':
            filename = self.iteration_to_file[iteration]
            return h5py_reader.read_species_data(
                    filename, iteration, species, record_comp, extensions )
        elif self.backend == 'openpmd-api':
            return io_reader.read_species_data(
                    self.series, iteration, species, record_comp, extensions )

    def get_grid_parameters(self, iteration, avail_fields, metadata ):
        """
        Return the parameters of the spatial grid (grid size and grid range)
        in two dictionaries

        Parameters:
        -----------
        iteration: int
            The iteration at which to extract the parameters

        avail_fields: list
           A list of the available fields
           e.g. ['B', 'E', 'rho']

        metadata: dictionary
          A dictionary whose keys are the fields of `avail_fields` and
          whose values are dictionaries that contain metadata (e.g. geometry)

        Returns:
        --------
        A tuple with `grid_size_dict` and `grid_range_dict`
        Both objects are dictionaries, with their keys being the labels of
        the axis of the grid (e.g. 'x', 'y', 'z')
        The values of `grid_size_dict` are the number of gridpoints along
        each axis.
        The values of `grid_range_dict` are lists of two floats, which
        correspond to the min and max of the grid, along each axis.
        """
        if self.backend == 'h5py':
            filename = self.iteration_to_file[iteration]
            return h5py_reader.get_grid_parameters(
                filename, iteration, avail_fields, metadata )
        elif self.backend == 'openpmd-api':
            return io_reader.get_grid_parameters(
                self.series, iteration, avail_fields, metadata )



"""
This file is part of the openPMD-viewer.

It defines an ordering Enum to allow for flexible ordering of components

Copyright 2022, openPMD-viewer contributors
Authors: Axel Huebl, Ryan Sandberg
License: 3-Clause-BSD-LBNL
"""

import enum
 
# creating enumerations using class
class RZorder(enum.Enum):
    """The index names of RZ axes in C order"""
    mrz = 1  # z is the fastest varying index in memory
    mzr = 2  # r is the fastest varying index in memory


order_error_msg = 'Data order is unupported. '
order_error_msg += 'Allowed orderings: '
order_error_msg += ' '.join([val.name for val in RZorder])



"""
openPMD-viewer

Usage
-----
See the class OpenPMDTimeSeries to open a set of openPMD files
"""
# Make the OpenPMDTimeSeries object accessible from outside the package
from .openpmd_timeseries import OpenPMDTimeSeries, FieldMetaInformation, \
    ParticleTracker

# Define the version number
from .__version__ import __version__
__all__ = ['OpenPMDTimeSeries', 'FieldMetaInformation',
           'ParticleTracker', '__version__']



"""
This file is part of the openPMD-viewer.

It defines the main FieldMetaInformation class, which
is returned by `get_field` along with the array of field values,
and gathers information collected from the openPMD file.

Copyright 2015-2016, openPMD-viewer contributors
Author: Remi Lehe
License: 3-Clause-BSD-LBNL
"""

import numpy as np


class FieldMetaInformation(object):
    """
    An object that is typically returned along with an array of field
    values, and which contains meta-information about the grid.

    Attributes
    ----------
    - axes: dict
        A dictionary of the form {0:'x', 1:'z'}, which indicates the name
        of the coordinate along each axis of the field array.
        For instance, in the case of {0:'x', 1:'y'}, the first axis of field
        array corresponds to the 'x' coordinate, and the second to 'y'.

    - xmin, xmax, zmin, zmax: double
        Scalars that indicate the position of the first grid point and
        last grid point along each axis.
        Notice that the name of these variables change according to
        the values in `axes`. For instance, if `axes` is {0: 'x', 1: 'y'},
        then these variables will be called xmin, xmax, ymin, ymax.

    - dx, dz: double
        Scalars that indicate the resolution of the grid on each axis.
        Notice that the name of these variables change according to
        the values in `axes`. For instance, if `axes` is {0: 'x', 1: 'y'},
        then these variables will be called dx and dy.

     - x, z: 1darrays of double
        The position of all the gridpoints, along each axis
        Notice that the name of these variables change according to
        the values in `axes`. For instance, if `axes` is {0: 'x', 1: 'y'},
        then these variables will be called x, y.

    - imshow_extent: 1darray
        (Only for 2D data)
        An array of 4 elements that can be passed as the `extent` in
        matplotlib's imshow function.
        Because of the API of the imshow function, the coordinates are
        'swapped' inside imshow_extent. For instance, if axes is
        {0: 'x', 1: 'y'}, then imshow_extent will be [ymin, ymax, xmin, xmax].

        (NB: in the details, imshow_extent contains slightly different values
        than ymin, ymax, xmin, xmax: these values are shifted by half a cell.
        The reason for this is that imshow plots a finite-width square for each
        value of the field array.)

    - t: float (in seconds), optional
        The simulation time of the data
        It allows the user to get the simulation time when calling the
        get_particle method for a given iteration and vice versa.
        Either `t` or `iteration` should be given.

    - iteration: int
        The iteration of the data
        It allows the user to get the simulation time when calling the
        get_particle method for a given iteration and vice versa.
        Either `t` or `iteration` should be given.

    """

    def __init__(self, axes, shape, grid_spacing,
                 global_offset, grid_unitSI, position, t, iteration, thetaMode=False):
        """
        Create a FieldMetaInformation object

        The input arguments correspond to their openPMD standard definition
        """
        # Register important initial information
        self.axes = axes

        # Create the elements
        for axis in sorted(axes.keys()):
            # Create the coordinates along this axis
            step = grid_spacing[axis] * grid_unitSI
            n_points = shape[axis]
            start = global_offset[axis] * grid_unitSI + position[axis] * step
            end = start + (n_points - 1) * step
            axis_points = np.linspace(start, end, n_points, endpoint=True)
            # Create the points below the axis if thetaMode is true
            if axes[axis] == 'r' and thetaMode:
                axis_points = np.concatenate((-axis_points[::-1], axis_points))
                start = -end
            # Register the results in the object
            axis_name = axes[axis]
            setattr(self, axis_name, axis_points)
            setattr(self, 'd' + axis_name, step)
            setattr(self, axis_name + 'min', axis_points[0])
            setattr(self, axis_name + 'max', axis_points[-1])

        # Register current simulation time and iteration in the object
        setattr(self, 'time', t)
        setattr(self, 'iteration', iteration)

        self._generate_imshow_extent()


    def restrict_to_1Daxis(self, axis):
        """
        Suppresses the information that correspond to other axes than `axis`

        Parameters
        ----------
        axis: string
            The axis to keep
            This has to be one of the keys of the self.axes dictionary
        """
        # Check if axis is a valid key
        if (axis in self.axes.values()) is False:
            raise ValueError('`axis` is not one of the coordinates '
                             'that are present in this object.')

        # Loop through the coordinates and suppress them
        for obsolete_axis in list(self.axes.values()):
            if obsolete_axis != axis:
                self._remove_axis(obsolete_axis)


    def _generate_imshow_extent(self):
        """
        Generate the list `imshow_extent`, which can be used directly
        as the argument `extent` of matplotlib's `imshow` command
        """
        if len(self.axes) == 2:
            self.imshow_extent = []
            for label in [self.axes[1], self.axes[0]]:
                coord_min = getattr( self, label+'min' )
                coord_max = getattr( self, label+'max' )
                coord_step = getattr( self, 'd'+label )
                self.imshow_extent += [ coord_min - 0.5*coord_step,
                                   coord_max + 0.5*coord_step ]
            self.imshow_extent = np.array(self.imshow_extent)
        else:
            if hasattr(self, 'imshow_extent'):
                delattr(self, 'imshow_extent')


    def _remove_axis(self, obsolete_axis):
        """
        Remove the axis `obsolete_axis` from the MetaInformation object
        """
        delattr(self, obsolete_axis)
        delattr(self, obsolete_axis + 'min')
        delattr(self, obsolete_axis + 'max')
        # Rebuild the dictionary `axes`, by including the axis
        # label in the same order, but omitting obsolete_axis
        ndim = len(self.axes)
        self.axes = dict( enumerate([
            self.axes[i] for i in range(ndim) \
            if self.axes[i] != obsolete_axis ]))

        self._generate_imshow_extent()


    def _convert_cylindrical_to_3Dcartesian(self):
        """
        Convert FieldMetaInformation from cylindrical to 3D Cartesian
        """

        try:
            assert (self.axes[0] == 'r' and self.axes[1] == 'z') or (self.axes[0] == 'z' and self.axes[1] == 'r')
        except (KeyError, AssertionError):
            raise ValueError('_convert_cylindrical_to_3Dcartesian'
                ' can only be applied to a timeseries in thetaMode geometry')

        # Create x and y arrays
        self.x = self.r.copy()
        self.y = self.r.copy()
        del self.r

        # Create dx and dy
        self.dx = self.dr
        self.dy = self.dr
        del self.dr

        # Create xmin, xmax, ymin, ymax
        self.xmin = self.rmin
        self.ymin = self.rmin
        del self.rmin
        self.xmax = self.rmax
        self.ymax = self.rmax
        del self.rmax

        # Change axes
        self.axes = {0:'x', 1:'y', 2:'z'}



"""
This file is part of the openPMD-viewer.

It defines a set of helper data and functions which
are used by the other files.

Copyright 2015-2016, openPMD-viewer contributors
Authors: Remi Lehe, Axel Huebl
License: 3-Clause-BSD-LBNL
"""
import numpy as np


def chunk_to_slice(chunk):
    """
    Convert an openPMD_api.ChunkInfo to slice
    """
    stops = [a + b for a, b in zip(chunk.offset, chunk.extent)]
    indices_per_dim = zip(chunk.offset, stops)
    index_tuple = map(lambda s: slice(s[0], s[1], None), indices_per_dim)
    return tuple(index_tuple)


def get_data(series, record_component, i_slice=None, pos_slice=None,
             output_type=None):
    """
    Extract the data from a (possibly constant) dataset
    Slice the data according to the parameters i_slice and pos_slice

    Parameters:
    -----------
    series: openpmd_api.Series
        An open, readable openPMD-api series object

    record_component: an openPMD.Record_Component

    pos_slice: int or list of int, optional
        Slice direction(s).
        When None, no slicing is performed

    i_slice: int or list of int, optional
       Indices of slices to be taken.

    output_type: a numpy type
       The type to which the returned array should be converted

    Returns:
    --------
    An np.ndarray (non-constant dataset) or a single double (constant dataset)
    """
    # For back-compatibility: Convert pos_slice and i_slice to
    # single-element lists if they are not lists (e.g. float
    # and int respectively).
    if pos_slice is not None and not isinstance(pos_slice, list):
        pos_slice = [pos_slice]
    if i_slice is not None and not isinstance(i_slice, list):
        i_slice = [i_slice]

    # ADIOS2: Actual chunks, all other: one chunk
    chunks = record_component.available_chunks()

    # read whole data set
    if pos_slice is None:
        # mask invalid regions with NaN
        #   note: full_like triggers a full read, thus we avoid it #340
        data = np.full(record_component.shape, np.nan, record_component.dtype)
        for chunk in chunks:
            chunk_slice = chunk_to_slice(chunk)

            # skip empty slices
            # https://github.com/ornladios/ADIOS2
            volume = 1
            for csl in chunk_slice:
                volume *= csl.stop - csl.start
            if volume == 0:
                continue

            # read only valid region
            x = record_component[chunk_slice]
            series.flush()
            data[chunk_slice] = x
    # slice: read only part of the data set
    else:
        full_shape = record_component.shape

        slice_shape = list(full_shape)      # copy
        pos_slice_sorted = pos_slice.copy() # copy for in-place sort
        pos_slice_sorted.sort(reverse=True)
        for dir_index in pos_slice_sorted:  # remove indices in list
            del slice_shape[dir_index]

        # mask invalid regions with NaN
        data = np.full(slice_shape, np.nan, dtype=record_component.dtype)

        # build requested ND slice with respect to full data
        s = []
        for d in range(len(full_shape)):
            if d in pos_slice:
                s.append(i_slice[pos_slice.index(d)]) # one index in such directions
            else: # all indices in other direction
                s.append(slice(None, None, None))
        s = tuple(s)

        # now we check which chunks contribute to the slice
        for chunk in chunks:
            skip_this_chunk = False
            s_valid = list(s)  # same as s but reduced to valid regions in chunk
            s_target = []  # starts and stops in sliced array
            chunk_slice = chunk_to_slice(chunk)

            # skip empty slices
            # https://github.com/ornladios/ADIOS2
            volume = 1
            for csl in chunk_slice:
                volume *= csl.stop - csl.start
            if volume == 0:
                continue

            # read only valid region
            for d, slice_d in enumerate(s):
                start = chunk_slice[d].start
                stop = chunk_slice[d].stop
                if isinstance(slice_d, int):
                    # Nothing to do for s_target (dimension sliced out)
                    # Nothing to do for s_valid (dimension index is set)
                    if slice_d < start or slice_d >= stop:
                        # chunk not in slice line/plane
                        skip_this_chunk = True
                else:
                    if slice_d.start is None or slice_d.start < start:
                        s_valid[d] = slice(start, s_valid[d].stop)
                    if slice_d.stop is None or slice_d.stop > stop:
                        s_valid[d] = slice(s_valid[d].start, stop)
                    s_target.append(slice(start, stop))

            s_valid = tuple(s_valid)
            s_target = tuple(s_target)

            # read
            if not skip_this_chunk:
                x = record_component[s_valid]
                series.flush()
                data[s_target] = x

    # Convert to the right type
    if (output_type is not None) and (data.dtype != output_type):
        data = data.astype( output_type )
    # Scale by the conversion factor
    if record_component.unit_SI != 1.0:
        if np.issubdtype(data.dtype, np.floating) or \
            np.issubdtype(data.dtype, np.complexfloating):
            data *= record_component.unit_SI
        else:
            data = data * record_component.unit_SI

    return data


def join_infile_path(*paths):
    """
    Join path components using '/' as separator.
    This method is defined as an alternative to os.path.join, which uses '\\'
    as separator in Windows environments and is therefore not valid to navigate
    within data files.

    Parameters:
    -----------
    *paths: all strings with path components to join

    Returns:
    --------
    A string with the complete path using '/' as separator.
    """
    # Join path components
    path = '/'.join(paths)
    # Correct double slashes, if any is present
    path = path.replace('//', '/')

    return path



"""
This file is part of the openPMD-viewer.

It defines a set of methods which are useful for plotting
(and labeling the plots).

Copyright 2015-2016, openPMD-viewer contributors
Author: Remi Lehe
License: 3-Clause-BSD-LBNL
"""
import numpy as np
import math
try:
    import warnings
    import matplotlib
    import matplotlib.pyplot as plt
    matplotlib_installed = True
except ImportError:
    matplotlib_installed = False

from .numba_wrapper import numba_installed
if numba_installed:
    from .utilities import histogram_cic_1d, histogram_cic_2d

# Redefine the default matplotlib formatter for ticks
if matplotlib_installed:
    from matplotlib.ticker import ScalarFormatter

    class PowerOfThreeFormatter( ScalarFormatter ):
        """
        Formatter for matplotlib's axes ticks,
        that prints numbers as e.g. 1.5e3, 3.2e6, 0.2e-9,
        where the exponent is always a multiple of 3.

        This helps a human reader to quickly identify the closest units
        (e.g. nanometer) of the plotted quantity.

        This class derives from `ScalarFormatter`, which
        provides a nice `offset` feature.
        """
        def __init__( self, *args, **kwargs ):
            ScalarFormatter.__init__( self, *args, **kwargs )
            # Do not print the order of magnitude on the side of the axis
            self.set_scientific(False)
            # Reduce the threshold for printing an offset on side of the axis
            self._offset_threshold = 2

        def __call__(self, x, pos=None):
            """
            Function called for each tick of an axis (for matplotlib>=3.1)
            Returns the string that appears in the plot.
            """
            return self.pprint_val( x, pos )

        def pprint_val( self, x, pos=None):
            """
            Function called for each tick of an axis (for matplotlib<3.1)
            Returns the string that appears in the plot.
            """
            # Calculate the exponent (power of 3)
            xp = (x - self.offset)
            if xp != 0:
                exponent = int(3 * math.floor( math.log10(abs(xp)) / 3 ))
            else:
                exponent = 0
            # Show 3 digits at most after decimal point
            mantissa = round( xp * 10**(-exponent), 3)
            # After rounding the exponent might change (e.g. 0.999 -> 1.)
            if mantissa != 0 and math.log10(abs(mantissa)) == 3:
                exponent += 3
                mantissa /= 1000
            string = "{:.3f}".format( mantissa )
            if '.' in string:
                # Remove trailing zeros and ., for integer mantissa
                string = string.rstrip('0')
                string = string.rstrip('.')
            if exponent != 0:
                string += "e{:d}".format( exponent )
            return string

    tick_formatter = PowerOfThreeFormatter()


class Plotter(object):

    """
    Class which is used for plotting particles and fields
    (and labeling the plots)
    """

    def __init__(self, t, iterations):
        """
        Initialize the object

        Parameters
        ----------
        t: 1darray of floats (seconds)
           Time for each available iteration of the timeseries

        iterations: 1darray of ints
           Iteration number for each available iteration of the timeseries
        """
        # Default fontsize
        self.fontsize = 12

        # Register the time array and iterations array
        # (Useful when labeling the figures)
        self.t = t
        self.iterations = iterations

    def hist1d(self, q1, w, quantity1, species, current_i, nbins, hist_range,
               cmap='Blues', vmin=None, vmax=None, deposition='cic', **kw):
        """
        Plot a 1D histogram of the particle quantity q1
        Sets the proper labels

        Parameters
        ----------
        q1: 1darray of floats
            An array with one element per macroparticle, representing
            the quantity to be plotted.

        w: 1darray of floats
            An array with one element per macroparticle, representing
            the number of real particles that correspond to each macroparticle

        quantity1: string
            The name of the quantity to be plotted (for labeling purposes)

        species: string
            The name of the species from which the data is taken

        current_i: int
            The index of this iteration, within the iterations list

        nbins : int
           Number of bins for the histograms

        hist_range : list contains 2 lists of 2 floats
           Extent of the histogram along each direction

        deposition : string
            Either `ngp` (Nearest Grid Point) or `cic` (Cloud-In-Cell)
            When plotting the particle histogram, this determines how
            particles affects neighboring bins.
            `cic` (which is the default) leads to smoother results than `ngp`.

        **kw : dict, otional
           Additional options to be passed to matplotlib's bar function
        """
        # Check if matplotlib is available
        check_matplotlib()

        # Find the iteration and time
        iteration = self.iterations[current_i]
        time = self.t[current_i]

        # Check deposition method
        if deposition == 'cic' and not numba_installed:
            print_cic_unavailable()
            deposition = 'ngp'

        # Bin the particle data
        q1 = q1.astype( np.float64 )
        if deposition == 'ngp':
            binned_data, _ = np.histogram(q1, nbins, hist_range[0], weights=w)
        elif deposition == 'cic':
            binned_data = histogram_cic_1d(
                q1, w, nbins, hist_range[0][0], hist_range[0][1])
        else:
            raise ValueError('Unknown deposition method: %s' % deposition)

        # Do the plot
        bin_size = (hist_range[0][1] - hist_range[0][0]) / nbins
        bin_coords = hist_range[0][0] + bin_size * ( 0.5 + np.arange(nbins) )
        plt.bar( bin_coords, binned_data, width=bin_size, **kw )
        plt.xlim( hist_range[0] )
        plt.ylim( hist_range[1] )
        plt.xlabel(quantity1, fontsize=self.fontsize)
        plt.title("%s:   t =  %.2e s    (iteration %d)"
                  % (species, time, iteration), fontsize=self.fontsize)
        # Format the ticks
        ax = plt.gca()
        ax.get_xaxis().set_major_formatter( tick_formatter )
        ax.get_yaxis().set_major_formatter( tick_formatter )

    def hist2d(self, q1, q2, w, quantity1, quantity2, species, current_i,
                nbins, hist_range, cmap='Blues', vmin=None, vmax=None,
                deposition='cic', **kw):
        """
        Plot a 2D histogram of the particle quantity q1
        Sets the proper labels

        Parameters
        ----------
        q1: 1darray of floats
            An array with one element per macroparticle, representing
            the quantity to be plotted.

        w: 1darray of floats
            An array with one element per macroparticle, representing
            the number of real particles that correspond to each macroparticle

        quantity1, quantity2: strings
            The name of the quantity to be plotted (for labeling purposes)

        species: string
            The name of the species from which the data is taken

        current_i: int
            The index of this iteration, within the iterations list

        nbins : list of 2 ints
           Number of bins along each direction, for the histograms

        hist_range : list contains 2 lists of 2 floats
           Extent of the histogram along each direction

        deposition : string
            Either `ngp` (Nearest Grid Point) or `cic` (Cloud-In-Cell)
            When plotting the particle histogram, this determines how
            particles affects neighboring bins.
            `cic` (which is the default) leads to smoother results than `ngp`.

        **kw : dict, otional
           Additional options to be passed to matplotlib's imshow function
        """
        # Check if matplotlib is available
        check_matplotlib()

        # Find the iteration and time
        iteration = self.iterations[current_i]
        time = self.t[current_i]

        # Check deposition method
        if deposition == 'cic' and not numba_installed:
            print_cic_unavailable()
            deposition = 'ngp'

        # Bin the particle data
        q1 = q1.astype( np.float64 )
        q2 = q2.astype( np.float64 )
        if deposition == 'ngp':
            binned_data, _, _ = np.histogram2d(
                q1, q2, nbins, hist_range, weights=w)
        elif deposition == 'cic':
            binned_data = histogram_cic_2d( q1, q2, w,
                nbins[0], hist_range[0][0], hist_range[0][1],
                nbins[1], hist_range[1][0], hist_range[1][1] )
        else:
            raise ValueError('Unknown deposition method: %s' % deposition)

        # Do the plot
        plt.imshow( binned_data.T, extent=hist_range[0] + hist_range[1],
             origin='lower', interpolation='nearest', aspect='auto',
             cmap=cmap, vmin=vmin, vmax=vmax, **kw )
        plt.colorbar()
        plt.xlabel(quantity1, fontsize=self.fontsize)
        plt.ylabel(quantity2, fontsize=self.fontsize)
        plt.title("%s:   t =  %.2e s   (iteration %d)"
                  % (species, time, iteration), fontsize=self.fontsize)
        # Format the ticks
        ax = plt.gca()
        ax.get_xaxis().set_major_formatter( tick_formatter )
        ax.get_yaxis().set_major_formatter( tick_formatter )

    def show_field_1d( self, F, info, field_label, current_i, plot_range,
                            vmin=None, vmax=None, **kw ):
        """
        Plot the given field in 1D

        Parameters
        ----------
        F: 1darray of floats
            Contains the field to be plotted

        info: a FieldMetaInformation object
            Contains the information about the plotted field

        field_label: string
           The name of the field plotted (for labeling purposes)

        vmin, vmax: floats or None
           The amplitude of the field

        plot_range : list of lists
           Indicates the values between which to clip the plot,
           along the 1st axis (first list) and 2nd axis (second list)
        """
        # Check if matplotlib is available
        check_matplotlib()

        # Find the iteration and time
        iteration = self.iterations[current_i]
        time = self.t[current_i]

        # Get the x axis
        xaxis = getattr( info, info.axes[0] )
        # Plot the data
        if np.issubdtype(F.dtype, np.complexfloating):
            plot_data = abs(F) # For complex numbers, plot the absolute value
            title = "|%s|" %field_label
        else:
            plot_data = F
            title = "%s" %field_label

        # Get the title and labels
        title += " at %.2e s   (iteration %d)" % (time, iteration)
        plt.title(title, fontsize=self.fontsize)
        # Add the name of the axes
        plt.xlabel('$%s \;(m)$' % info.axes[0], fontsize=self.fontsize)

        plt.plot( xaxis, plot_data )
        # Get the limits of the plot
        # - Along the first dimension
        if (plot_range[0][0] is not None) and (plot_range[0][1] is not None):
            plt.xlim( plot_range[0][0], plot_range[0][1] )
        else:
            plt.xlim( xaxis.min(), xaxis.max() )  # Full extent of the box
        # - Along the second dimension
        if (plot_range[1][0] is not None) and (plot_range[1][1] is not None):
            plt.ylim( plot_range[1][0], plot_range[1][1] )
        # Format the ticks
        ax = plt.gca()
        ax.get_xaxis().set_major_formatter( tick_formatter )
        ax.get_yaxis().set_major_formatter( tick_formatter )

    def show_field_2d(self, F, info, slice_across, m, field_label, geometry,
                        current_i, plot_range, **kw):
        """
        Plot the given field in 2D

        Parameters
        ----------
        F: 2darray of floats
            Contains the field to be plotted

        info: a FieldMetaInformation object
            Contains the information about the plotted field

        slice_across : str, optional
           Only used for 3dcartesian geometry
           The direction across which the data is sliced

        m: int
           Only used for thetaMode geometry
           The azimuthal mode used when plotting the fields

        field_label: string
           The name of the field plotted (for labeling purposes)

        geometry: string
           Either "2dcartesian", "3dcartesian" or "thetaMode"

        plot_range : list of lists
           Indicates the values between which to clip the plot,
           along the 1st axis (first list) and 2nd axis (second list)
        """
        # Check if matplotlib is available
        check_matplotlib()

        # Find the iteration and time
        iteration = self.iterations[current_i]
        time = self.t[current_i]

        # Plot the data
        if np.issubdtype(F.dtype, np.complexfloating):
            plot_data = abs(F)
            title = "|%s|" %field_label
        else:
            plot_data = F
            title = "%s" %field_label
        plt.imshow(plot_data, extent=info.imshow_extent, origin='lower',
                   interpolation='nearest', aspect='auto', **kw)
        plt.colorbar()

        # Get the title and labels
        # Cylindrical geometry
        if geometry == "thetaMode":
            mode = str(m)
            title += " in the mode %s at %.2e s   (iteration %d)" \
                      % (mode, time, iteration)
        # 2D Cartesian geometry
        else:
            title += " at %.2e s   (iteration %d)" % (time, iteration)
        plt.title(title, fontsize=self.fontsize)

        # Add the name of the axes
        plt.xlabel('$%s \;(m)$' % info.axes[1], fontsize=self.fontsize)
        plt.ylabel('$%s \;(m)$' % info.axes[0], fontsize=self.fontsize)

        # Get the limits of the plot
        # - Along the first dimension
        if (plot_range[0][0] is not None) and (plot_range[0][1] is not None):
            plt.xlim( plot_range[0][0], plot_range[0][1] )
        # - Along the second dimension
        if (plot_range[1][0] is not None) and (plot_range[1][1] is not None):
            plt.ylim( plot_range[1][0], plot_range[1][1] )
        # Format the ticks
        ax = plt.gca()
        ax.get_xaxis().set_major_formatter( tick_formatter )
        ax.get_yaxis().set_major_formatter( tick_formatter )


def print_cic_unavailable():
    warnings.warn(
        "\nCIC particle histogramming is unavailable because \n"
        "Numba is not installed. NGP histogramming is used instead.\n"
        "Please considering installing numba (e.g. `pip install numba`)")


def check_matplotlib():
    """Raise error messages or warnings when potential issues when
    potenial issues with matplotlib are detected."""

    if not matplotlib_installed:
        raise RuntimeError( "Failed to import the openPMD-viewer plotter.\n"
            "(Make sure that matplotlib is installed.)")

    elif ('MacOSX' in matplotlib.get_backend()):
        warnings.warn("\n\nIt seems that you are using the matplotlib MacOSX "
        "backend. \n(This typically obtained when typing `%matplotlib`.)\n"
        "With recent version of Jupyter, the plots might not appear.\nIn this "
        "case, switch to `%matplotlib notebook` and restart the notebook.")



"""
This is a stub that detects whether the user is attempting to use the
old import statement from openPMD-viewer 0.X (`import opmd_viewer`).

In that case, an exception is raised that prompts the user to use the new API.
"""
# Define the version number
from openpmd_viewer import __version__
__all__ = ['__version__']

raise DeprecationWarning("""
It looks like you are trying to use the API from openPMD-viewer version 0.X
but the installed version on your system is openPMD-viewer version 1.X.

* If you wish to use the new openPMD-viewer version 1.X, the import statement
should use `openpmd_viewer` instead of `opmd_viewer`, e.g.:
```
from openpmd_viewer import OpenPMDTimeSeries
```
Please have a look at the list of the changes introduced in version 1.X here:
https://github.com/openPMD/openPMD-viewer/blob/upcoming-1.0/CHANGELOG.md#10
In particular, note that `get_particle` now returns particle positions in
meters (not in microns anymore) and that the syntax for slicing fields has
changed.

* If you wish to go back to the old openPMD-viewer version 0.X, use:
```
pip install openPMD-viewer==0.9
```
""")



import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

# Get the long description
with open('./README.md') as f:
    long_description = f.read()
# Get the package requirements from the requirements.txt file
with open('./requirements.txt') as f:
    install_requires = [line.strip('\n') for line in f.readlines()]

# Read the version number, by executing the file openpmd_viewer/__version__.py
# This defines the variable __version__
with open('./openpmd_viewer/__version__.py') as f:
    exec( f.read() )

# Define a custom class to run the py.test with `python setup.py test`
class PyTest(TestCommand):

    def run_tests(self):
        import pytest
        errcode = pytest.main([])
        sys.exit(errcode)

# Main setup command
setup(name='openPMD-viewer',
      version=__version__,
      description='Visualization tools for openPMD files',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/openPMD/openPMD-viewer.git',
      maintainer='Remi Lehe',
      maintainer_email='remi.lehe@lbl.gov',
      license='BSD-3-Clause-LBNL',
      packages=find_packages('.'),
      package_data={'openpmd_viewer': ['notebook_starter/*.ipynb']},
      scripts=['openpmd_viewer/notebook_starter/openPMD_notebook'],
      tests_require=['pytest', 'jupyter'],
      install_requires=install_requires,
      extras_require = {
        'all': ["ipympl", "ipywidgets", "matplotlib", "numba", "openpmd-api", "wget"],
        'GUI':  ["ipywidgets", "ipympl", "matplotlib"],
        'plot': ["matplotlib"],
        'tutorials': ["ipywidgets", "ipympl", "matplotlib", "wget"],
        'numba': ["numba"],
        'openpmd-api': ["openpmd-api"]
        },
      cmdclass={'test': PyTest},
      platforms='any',
      python_requires='>=3.8',
      classifiers=[
          'Programming Language :: Python',
          'Development Status :: 4 - Beta',
          'Natural Language :: English',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'Operating System :: OS Independent',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Scientific/Engineering :: Visualization',
          'Topic :: Database :: Front-Ends',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11'],
      )



from .particle_reader import read_species_data
from .params_reader import read_openPMD_params
from .field_reader import read_field_cartesian, \
    read_field_circ, get_grid_parameters

__all__ = ['read_species_data', 'read_openPMD_params', 'read_field_cartesian',
           'read_field_circ', 'get_grid_parameters']



"""
This file is part of the openPMD-viewer.

It defines a function that can read standard parameters from an openPMD file.

Copyright 2015-2016, openPMD-viewer contributors
Authors: Remi Lehe, Axel Huebl
License: 3-Clause-BSD-LBNL
"""

import h5py
import numpy as np
from .utilities import is_scalar_record, get_shape, join_infile_path


def read_openPMD_params(filename, iteration, extract_parameters=True):
    """
    Extract the time and some openPMD parameters from a file

    Parameter
    ---------
    filename: string
        The path to the file from which parameters should be extracted

    iteration : int
        The iteration at which to obtain the data

    extract_parameters: bool, optional
        Whether to extract all parameters or only the time
        (Function execution is faster when extract_parameters is False)

    Returns
    -------
    A tuple with:
    - A float corresponding to the time of this iteration in SI units
    - A dictionary containing several parameters, such as the geometry, etc.
      When extract_parameters is False, the second argument returned is None.
    """
    # Open the file, and do a version check
    f = h5py.File(filename, 'r')
    version = f.attrs['openPMD'].decode()
    if version[:2] != '1.':
        raise ValueError(
            "File %s is not supported: Invalid openPMD version: "
            "%s)" % (filename, version))

    # Find the base path object, and extract the time
    bpath = f['/data/{0}'.format( iteration )]
    t = bpath.attrs["time"] * bpath.attrs["timeUnitSI"]

    # If the user did not request more parameters, close file and exit
    if not extract_parameters:
        f.close()
        return(t, None)

    # Otherwise, extract the rest of the parameters
    params = {}

    # Find out supported openPMD extensions claimed by this file
    # note: a file might implement multiple extensions
    known_extensions = {'ED-PIC': np.uint32(1)}
    bitmask_all_extensions = f.attrs['openPMDextension']
    params['extensions'] = []
    for extension, bitmask in known_extensions.items():
        # This uses a bitmask to identify activated extensions
        # efficiently in static programming languages via
        # a single attribute and a binary AND (&) operation.
        # Standard: https://git.io/vwnMw
        # Bitmasks: https://en.wikipedia.org/wiki/Mask_%28computing%29
        if bitmask_all_extensions & bitmask == bitmask:
            params['extensions'].append(extension)

    # Find out whether fields are present and extract their metadata
    fields_available = False
    if ('meshesPath' in f.attrs):        # Check for openPMD 1.1 files
        meshes_path = f.attrs['meshesPath'].decode().strip('/')
        if meshes_path in bpath.keys():  # Check for openPMD 1.0 files
            fields_available = True
    if fields_available:
        params['avail_fields'] = []
        params['fields_metadata'] = {}

        # Loop through the available fields
        for field_name in bpath[meshes_path].keys():
            field = bpath[join_infile_path(meshes_path, field_name)]
            metadata = {}
            metadata['geometry'] = field.attrs['geometry'].decode()
            metadata['axis_labels'] = [ coord.decode() for coord in
                                        field.attrs['axisLabels'] ]
            # Swap the order of the labels if the code that wrote the HDF5 file
            # was Fortran order (i.e. reverse order with respect to Python)
            if field.attrs['dataOrder'].decode() == 'F':
                metadata['axis_labels'] = metadata['axis_labels'][::-1]
            # Check whether the field is a vector or a scalar
            if is_scalar_record(field):
                metadata['type'] = 'scalar'
                components = []
            else:
                metadata['type'] = 'vector'
                components = list(field.keys())
            # Register available components
            metadata['avail_components'] = components
            # Check the number of modes
            if metadata['geometry'] == "thetaMode":
                if is_scalar_record(field):
                    Nm, _, _ = get_shape(field)
                else:
                    coord = list(field.keys())[0]
                    Nm, _, _ = get_shape(field[coord])
                metadata['avail_circ_modes'] = ['all'] + \
                    [str(m) for m in range(int(Nm / 2) + 1)]
            # Check if this a 1d, 2d or 3d Cartesian
            elif metadata['geometry'] == "cartesian":
                dim = len(metadata['axis_labels'])
                if dim == 1:
                    metadata['geometry'] = "1dcartesian"
                elif dim == 2:
                    metadata['geometry'] = "2dcartesian"
                elif dim == 3:
                    metadata['geometry'] = "3dcartesian"
                metadata['avail_circ_modes'] = []

            params['avail_fields'].append( field_name )
            params['fields_metadata'][field_name] = metadata

    else:
        params['avail_fields'] = None

    # Find out whether particles are present, and if yes of which species
    particles_available = False
    if ('particlesPath' in f.attrs):        # Check for openPMD 1.1 files
        particle_path = f.attrs['particlesPath'].decode().strip('/')
        if particle_path in bpath.keys():   # Check for openPMD 1.0 files
            # Check that there is at least one species
            if len(bpath[particle_path].keys()) > 0:
                particles_available = True
    if particles_available:
        # Particles are present ; extract the species
        params['avail_species'] = []
        for species_name in bpath[particle_path].keys():
            params['avail_species'].append(species_name)
        # dictionary with list of record components for each species
        record_components = {}
        # Go through all species
        for species_name in iter(params['avail_species']):
            species = bpath[join_infile_path(particle_path, species_name)]
            record_components[species_name] = []

            # Go through all the particle records of this species
            for record_name in species.keys():
                # Skip the particlePatches, which are not used here.
                if record_name == 'particlePatches':
                    continue
                record = species[record_name]
                if is_scalar_record(record):
                    # Add the name of the scalar record
                    record_components[species_name]. \
                        append(record_name)
                else:
                    # Add each component of the vector record
                    for coord in record.keys():
                        record_components[species_name]. \
                            append(join_infile_path(record_name, coord))
            # Simplify the name of some standard openPMD records
            record_components[species_name] = \
                simplify_record(record_components[species_name])
        params['avail_record_components'] = record_components
        # deprecated
        first_species_name = next(iter(params['avail_species']))
        params['avail_ptcl_quantities'] = \
            record_components[first_species_name]
    else:
        # Particles are absent
        params['avail_species'] = None
        params['avail_record_components'] = None
        # deprecated
        params['avail_ptcl_quantities'] = None

    # Close the file and return the parameters
    f.close()
    return(t, params)


def simplify_record(record_comps):
    """
    Replace the names of some standard record by shorter names

    Parameter
    ---------
    record_comps: a list of strings
        A list of available particle record components

    Returns
    -------
    A list with shorter names, where applicable
    """
    # Replace the names of the positions
    if ('position/x' in record_comps) and ('positionOffset/x' in record_comps):
        record_comps.remove('position/x')
        record_comps.remove('positionOffset/x')
        record_comps.append('x')
    if ('position/y' in record_comps) and ('positionOffset/y' in record_comps):
        record_comps.remove('position/y')
        record_comps.remove('positionOffset/y')
        record_comps.append('y')
    if ('position/z' in record_comps) and ('positionOffset/z' in record_comps):
        record_comps.remove('position/z')
        record_comps.remove('positionOffset/z')
        record_comps.append('z')

    # Replace the names of the momenta
    if 'momentum/x' in record_comps:
        record_comps.remove('momentum/x')
        record_comps.append('ux')
    if 'momentum/y' in record_comps:
        record_comps.remove('momentum/y')
        record_comps.append('uy')
    if 'momentum/z' in record_comps:
        record_comps.remove('momentum/z')
        record_comps.append('uz')

    # Replace the name for 'weights'
    if 'weighting' in record_comps:
        record_comps.remove('weighting')
        record_comps.append('w')

    return(record_comps)
